from typing import Any, List, Dict, Union
from orichain import error_explainer

import asyncio

VERSION = "3.4.1"


class Embed(object):
    """
    Synchronous Embed class to get embeddings from SentenceTransformers.
    """

    def __init__(self, **kwds: Any) -> None:
        """
        Loads SentenceTransformer model and initializes it.

        Args:
            - model_name (str): name of hf model to be loaded
            - model_download_path (str): path to download the model
            - device (str): device to be used for inference
            - trust_remote_code (bool): whether to trust remote code
            - token (str): token for downloading model

        Raises:
            - ImportError: If sentence-transformers is not installed, it raises ImportError.
        """

        # Check if sentence-transformers is installed
        try:
            from sentence_transformers import SentenceTransformer
        except ImportError:
            install = (
                input(
                    "sentence-transformers is not installed. Do you want to install it now? (y/n): "
                )
                .strip()
                .lower()
            )
            if install == "y" or install == "yes":
                import subprocess

                subprocess.run(
                    ["pip", "install", f"sentence-transformers=={VERSION}"], check=True
                )
            else:
                raise ImportError(
                    f"sentence-transformers is required for embeddings functionalities ({kwds.get('model_name', 'NA')}). Please install it manually using `pip install orichain[sentence-transformers]' or 'pip install sentence-transformers=={VERSION}`."
                )

        self.default_model_dir = kwds.get(
            "model_download_path", "/home/ubuntu/projects/models/embedding_models"
        )

        from sentence_transformers import SentenceTransformer
        import os

        # Check if model is already downloaded
        if not os.path.isdir(f"{self.default_model_dir}/{kwds.get('model_name')}"):
            self.model = SentenceTransformer(
                model_name_or_path=kwds.get("model_name"),
                device=kwds.get("device", "cpu"),
                trust_remote_code=kwds.get("trust_remote_code", False),
                token=kwds.get("token", None),
            )
            self.model.save(f"{self.default_model_dir}/{kwds.get('model_name')}")
        else:
            self.model = SentenceTransformer(
                model_name_or_path=f"{self.default_model_dir}/{kwds.get('model_name')}",
                device=kwds.get("device", "cpu"),
                trust_remote_code=kwds.get("trust_remote_code", False),
                token=kwds.get("token", None),
                local_files_only=True,
            )

    def __call__(
        self, text: Union[str, List[str]], **kwds: Any
    ) -> Union[List[float], List[List[float]], Dict]:
        """
        Get embeddings for the given text(s).

        Args:
            - text (Union[str, List[str]]): Input text or list of texts
            - **kwargs: Additional keyword arguments for the embedding API

        Returns:
            Union[List[float], List[List[float]], np.ndarray, torch.Tensor, Dict[str, Any]]: Embeddings in the requested format, or error information.
        """
        try:
            if isinstance(text, str):
                text = [text]

            embeddings = self.model.encode(
                sentences=text,
                prompt_name=kwds.get("prompt_name", None),
                prompt=kwds.get("prompt", None),
                output_value=kwds.get("output_value", "sentence_embedding"),
                show_progress_bar=kwds.get("show_progress_bar", False),
                precision=kwds.get("precision", "float32"),
                batch_size=kwds.get("batch_size", 32),
                convert_to_tensor=kwds.get("convert_to_tensor", False),
                convert_to_numpy=kwds.get("convert_to_numpy", False),
                device=kwds.get("device", None),
                normalize_embeddings=kwds.get("normalize_embeddings", False),
            )

            if not kwds.get("convert_to_tensor") and not kwds.get("convert_to_numpy"):
                embeddings = [embedding.tolist() for embedding in embeddings]

            if len(embeddings) == 1:
                embeddings = embeddings[0]

            return embeddings
        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}


class AsyncEmbed(object):
    """
    Asynchronous Embed class to get embeddings from SentenceTransformers.
    """

    def __init__(self, **kwds: Any) -> None:
        """
        Loads SentenceTransformer model and initializes it.

        Args:
            - model_name (str): name of hf model to be loaded
            - model_download_path (str): path to download the model
            - device (str): device to be used for inference
            - trust_remote_code (bool): whether to trust remote code
            - token (str): token for downloading model

        Raises:
            - ImportError: If sentence-transformers is not installed, it raises ImportError.
        """

        # Check if sentence-transformers is installed
        try:
            from sentence_transformers import SentenceTransformer
        except ImportError:
            install = (
                input(
                    "sentence-transformers is not installed. Do you want to install it now? (y/n): "
                )
                .strip()
                .lower()
            )
            if install == "y" or install == "yes":
                import subprocess

                subprocess.run(
                    ["pip", "install", f"sentence-transformers=={VERSION}"], check=True
                )
            else:
                raise ImportError(
                    f"sentence-transformers is required for embeddings functionalities ({kwds.get('model_name', 'NA')}). Please install it manually using `pip install orichain[sentence-transformers]' or 'pip install sentence-transformers=={VERSION}`."
                )

        self.default_model_dir = kwds.get(
            "model_download_path", "/home/ubuntu/projects/models/embedding_models"
        )

        from sentence_transformers import SentenceTransformer
        import os

        # Check if model is already downloaded
        if not os.path.isdir(f"{self.default_model_dir}/{kwds.get('model_name')}"):
            self.model = SentenceTransformer(
                model_name_or_path=kwds.get("model_name"),
                device=kwds.get("device", "cpu"),
                trust_remote_code=kwds.get("trust_remote_code", False),
                token=kwds.get("token", None),
            )
            self.model.save(f"{self.default_model_dir}/{kwds.get('model_name')}")
        else:
            self.model = SentenceTransformer(
                model_name_or_path=f"{self.default_model_dir}/{kwds.get('model_name')}",
                device=kwds.get("device", "cpu"),
                trust_remote_code=kwds.get("trust_remote_code", False),
                token=kwds.get("token", None),
                local_files_only=True,
            )

    async def __call__(
        self, text: Union[str, List[str]], **kwds: Any
    ) -> Union[List[float], List[List[float]], Dict]:
        """
        Get embeddings for the given text(s).

        Args:
            - text (Union[str, List[str]]): Input text or list of texts
            - **kwargs: Additional keyword arguments for the embedding API

        Returns:
            Union[List[float], List[List[float]], np.ndarray, torch.Tensor, Dict[str, Any]]: Embeddings in the requested format, or error information.
        """
        try:
            if isinstance(text, str):
                text = [text]

            embeddings = await asyncio.to_thread(
                self.model.encode,
                sentences=text,
                prompt_name=kwds.get("prompt_name", None),
                prompt=kwds.get("prompt", None),
                output_value=kwds.get("output_value", "sentence_embedding"),
                show_progress_bar=kwds.get("show_progress_bar", False),
                precision=kwds.get("precision", "float32"),
                batch_size=kwds.get("batch_size", 32),
                convert_to_tensor=kwds.get("convert_to_tensor", False),
                convert_to_numpy=kwds.get("convert_to_numpy", False),
                device=kwds.get("device", None),
                normalize_embeddings=kwds.get("normalize_embeddings", False),
            )

            if not kwds.get("convert_to_tensor") and not kwds.get("convert_to_numpy"):
                embeddings = [embedding.tolist() for embedding in embeddings]

            if len(embeddings) == 1:
                embeddings = embeddings[0]

            return embeddings
        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}



from typing import List, Optional, Dict
from orichain import error_explainer

import asyncio

VERSION = "2.1.0"


class LanguageDetection(object):
    """
    Synchronous interface for detecting the language of user messages using lingua-language-detector library.
    """

    def __init__(
        self,
        languages: Optional[List[str]] = None,
        min_words: Optional[int] = None,
        low_accuracy: Optional[bool] = False,
    ) -> None:
        """Loading detector with requirements, by default loads all the languages with 0.0 min confidence

        Args:
            - languages (Optional[List], optional): List of languages to load. Defaults to None.
            - min_words (Optional[int], optional): Minimum words in the user message to detect language. Defaults to None.
            - low_accuracy (Optional[bool], optional): To enable low accuracy mode. Defaults to False.
        """

        try:
            from lingua import Language, LanguageDetectorBuilder
        except ImportError:
            install = (
                input(
                    "lingua-language-detector is not installed. Do you want to install it now? (y/n): "
                )
                .strip()
                .lower()
            )
            if install == "y" or install == "yes":
                import subprocess

                subprocess.run(
                    ["pip", "install", f"lingua-language-detector=={VERSION}"],
                    check=True,
                )
            else:
                raise ImportError(
                    f"lingua-language-detector is required for LanguageDetection class. Please install it manually using `pip install orichain[lingua-language-detector]' or 'pip install lingua-language-detector=={VERSION}`."
                )

        try:
            # Loading detector with requirements
            if languages:
                # Loading detector with specific languages
                language_objects = [
                    getattr(Language, lang.upper()) for lang in languages
                ]
                detector = LanguageDetectorBuilder.from_languages(*language_objects)
            else:
                # Loading detector with all languages
                detector = LanguageDetectorBuilder.from_all_languages()

            # Enabling low accuracy mode
            if low_accuracy:
                detector = detector.with_low_accuracy_mode()

            # Building the detector
            self.detector = detector.with_preloaded_language_models().build()

            self.min_words = min_words
        except Exception as e:
            error_explainer(e)

    def __call__(
        self,
        user_message: str,
        min_words: Optional[int] = None,
        add_confidence: Optional[bool] = False,
        iso_code_639_3: Optional[bool] = False,
    ) -> Dict:
        """Runs language detection

        Args:
            - user_message (str): User message to detect language
            - min_words (Optional[int], optional): Minimum words in the user message to detect language. Defaults to None.
            - add_confidence (Optional[bool], optional): To add confidence in the result. Defaults to False.
            - iso_code_639_3 (Optional[bool], optional): To get iso code 639-3 instead of 639-1. Defaults to False.

        Returns:
            Dict: Result of language detection
        """

        try:
            result = {"user_lang": None}
            min_words = min_words or self.min_words
            if min_words:
                if len(user_message.split()) < min_words:
                    return result

            output = self.detector.compute_language_confidence_values(text=user_message)

            result["user_lang"] = (
                output[0].language.iso_code_639_1.name
                if not iso_code_639_3
                else output[0].language.iso_code_639_1.name
            )

            if add_confidence:
                result["confidence"] = output[0].value

            return result
        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}


class AsyncLanguageDetection(object):
    """
    Asynchronous interface for detecting the language of user messages using lingua-language-detector library.
    """

    def __init__(
        self,
        languages: Optional[List[str]] = None,
        min_words: Optional[int] = None,
        low_accuracy: Optional[bool] = False,
    ) -> None:
        """Loading detector with requirements, by default loads all the languages with 0.0 min confidence

        Args:
            - languages (Optional[List], optional): List of languages to load. Defaults to None.
            - min_words (Optional[int], optional): Minimum words in the user message to detect language. Defaults to None.
            - low_accuracy (Optional[bool], optional): To enable low accuracy mode. Defaults to False.
        """

        try:
            from lingua import Language, LanguageDetectorBuilder
        except ImportError:
            install = (
                input(
                    "lingua-language-detector is not installed. Do you want to install it now? (y/n): "
                )
                .strip()
                .lower()
            )
            if install == "y" or install == "yes":
                import subprocess

                subprocess.run(
                    ["pip", "install", f"lingua-language-detector=={VERSION}"],
                    check=True,
                )
            else:
                raise ImportError(
                    f"lingua-language-detector is required for AsyncLanguageDetection class. Please install it manually using `pip install orichain[lingua-language-detector]' or 'pip install lingua-language-detector=={VERSION}`."
                )

        try:
            if languages:
                language_objects = [
                    getattr(Language, lang.upper()) for lang in languages
                ]
                detector = LanguageDetectorBuilder.from_languages(*language_objects)
            else:
                detector = LanguageDetectorBuilder.from_all_languages()
            if low_accuracy:
                detector = detector.with_low_accuracy_mode()

            self.detector = detector.with_preloaded_language_models().build()

            self.min_words = min_words
        except Exception as e:
            error_explainer(e)

    async def __call__(
        self,
        user_message: str,
        min_words: Optional[int] = None,
        add_confidence: Optional[bool] = False,
        iso_code_639_3: Optional[bool] = False,
    ) -> Dict:
        """Runs language detection

        Args:
            - user_message (str): User message to detect language
            - min_words (Optional[int], optional): Minimum words in the user message to detect language. Defaults to None.
            - add_confidence (Optional[bool], optional): To add confidence in the result. Defaults to False.
            - iso_code_639_3 (Optional[bool], optional): To get iso code 639-3 instead of 639-1. Defaults to False.

        Returns:
            Dict: Result of language detection
        """

        try:
            result = {"user_lang": None}
            min_words = min_words or self.min_words
            if min_words:
                if len(user_message.split()) < min_words:
                    return result

            output = await asyncio.to_thread(
                self.detector.compute_language_confidence_values, text=user_message
            )

            result["user_lang"] = (
                output[0].language.iso_code_639_1.name
                if not iso_code_639_3
                else output[0].language.iso_code_639_1.name
            )

            if add_confidence:
                result["confidence"] = output[0].value

            return result
        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}



from typing import (
    Any,
    Dict,
    List,
    Union,
)
from orichain import error_explainer


class Embed(object):
    """
    Synchronous Embed class to get embeddings from Gemini.
    """

    def __init__(self, **kwds) -> None:
        """
        Initialize Google client for Gemini and set up API key.

        Args:
            - api_key (str): Gemini API key
            - http_options (types.HttpOptions, optional): HTTP options to be used in each of the requests. Default is None
            - debug_config (DebugConfig, optional): Configuration options that change client network behavior when testing. Default is None

        Raises:
            - KeyError: If required parameters are not provided.
            - TypeError: If an invalid type is provided for a parameter
        """

        from google.genai import Client, types
        from google.genai.client import DebugConfig

        # Validate input parameters
        if not kwds.get("api_key"):
            raise KeyError("Required `api_key` not found")
        elif kwds.get("http_options") and not isinstance(
            kwds.get("http_options"), types.HttpOptions
        ):
            raise TypeError(
                "Invalid 'http_options' type detected:",
                type(kwds.get("http_options")),
                ", Please enter valid http_options using:\n'from google.genai.types import HttpOptions'",
            )
        elif kwds.get("debug_config") and not isinstance(
            kwds.get("debug_config"), DebugConfig
        ):
            raise TypeError(
                "Invalid 'debug_config' type detected:",
                type(kwds.get("debug_config")),
                ", Please enter valid debug_config using:\n'from google.genai.client import DebugConfig'",
            )
        else:
            pass

        # Initialize the Google client with provided parameters
        self.client: Client = Client(
            api_key=kwds.get("api_key"),
            http_options=kwds.get("http_options", types.HttpOptions(timeout=7000)),
            debug_config=kwds.get("debug_config"),
        )

    def __call__(
        self, text: Union[str, List[str]], model_name: str, **kwds: Any
    ) -> Union[List[float], List[List[float]], Dict]:
        """
        Get embeddings for the given text(s).

        Args:
            - text (Union[str, List[str]]): Input text or list of texts
            - model_name (str): Name of the embedding model to use
            - **kwargs: Additional keyword arguments for the embedding API

        Returns:
            (Union[List[float], List[List[float]], Dict[str, Any]]): Embeddings or error information
        """
        try:
            response = self.client.models.embed_content(
                model=model_name,
                contents=text if isinstance(text, List) else [text],
                config=kwds.get("config"),
            )

            embeddings = [embedding.values for embedding in response.embeddings]
            if len(embeddings) == 1:
                embeddings = embeddings[-1]
            return embeddings

        except Exception as e:
            error_explainer(e=e)
            return {"error": 500, "reason": str(e)}


class AsyncEmbed(object):
    """
    Asynchronous wrapper for Google's API client for Gemini.
    """

    def __init__(self, **kwds) -> None:
        """
        Initialize Google client for Gemini and set up API key.

        Args:
            - api_key (str): Gemini API key
            - http_options (types.HttpOptions, optional): HTTP options to be used in each of the requests. Default is None
            - debug_config (DebugConfig, optional): Configuration options that change client network behavior when testing. Default is None

        Raises:
            - KeyError: If required parameters are not provided.
            - TypeError: If an invalid type is provided for a parameter
        """

        from google.genai import Client, types
        from google.genai.client import DebugConfig

        # Validate input parameters
        if not kwds.get("api_key"):
            raise KeyError("Required `api_key` not found")
        elif kwds.get("http_options") and not isinstance(
            kwds.get("http_options"), types.HttpOptions
        ):
            raise TypeError(
                "Invalid 'http_options' type detected:",
                type(kwds.get("http_options")),
                ", Please enter valid http_options using:\n'from google.genai.types import HttpOptions'",
            )
        elif kwds.get("debug_config") and not isinstance(
            kwds.get("debug_config"), DebugConfig
        ):
            raise TypeError(
                "Invalid 'debug_config' type detected:",
                type(kwds.get("debug_config")),
                ", Please enter valid debug_config using:\n'from google.genai.client import DebugConfig'",
            )
        else:
            pass

        # Initialize the Google client with provided parameters
        self.client: Client = Client(
            api_key=kwds.get("api_key"),
            http_options=kwds.get("http_options", types.HttpOptions(timeout=7000)),
            debug_config=kwds.get("debug_config"),
        )

    async def __call__(
        self, text: Union[str, List[str]], model_name: str, **kwds: Any
    ) -> Union[List[float], List[List[float]], Dict]:
        """
        Get embeddings for the given text(s).

        Args:
            - text (Union[str, List[str]]): Input text or list of texts
            - model_name (str): Name of the embedding model to use
            - **kwargs: Additional keyword arguments for the embedding API

        Returns:
            (Union[List[float], List[List[float]], Dict[str, Any]]): Embeddings or error information
        """
        try:
            response = await self.client.aio.models.embed_content(
                model=model_name,
                contents=text if isinstance(text, List) else [text],
                config=kwds.get("config"),
            )
            embeddings = [embedding.values for embedding in response.embeddings]
            if len(embeddings) == 1:
                embeddings = embeddings[-1]

            return embeddings

        except Exception as e:
            error_explainer(e=e)
            return {"error": 500, "reason": str(e)}



import json
from typing import Dict, List, Optional, Generator, AsyncGenerator
from fastapi import Request

from orichain import error_explainer


class Generate(object):
    """
    Synchronous wrapper for Azure OpenAI's API client.

    This class provides methods for generating responses from OpenAI's models
    both in streaming and non-streaming modes. It handles chat history formatting,
    error handling, and proper request configuration.
    """

    def __init__(self, **kwds) -> None:
        """
        Initialize Azure OpenAI client and set up API key.

        Args:
            - api_key (str): Azure OpenAI API key.
            - azure_endpoint (str): Azure OpenAI endpoint.
            - api_version (str): Azure OpenAI API version.
            - timeout (Timeout, optional): Request timeout parameter like connect, read, write. Default: 60.0, 5.0, 10.0, 2.0
            - max_retries (int, optional): Number of retries for the request. Default: 2

        Raises:
            - KeyError: If required parameters are not provided.
            - TypeError: If an invalid type is provided for a parameter
        """

        from httpx import Timeout

        # Validate input parameters
        if not kwds.get("api_key"):
            raise KeyError("Required `api_key` not found")
        elif not kwds.get("azure_endpoint"):
            raise KeyError("Required `azure_endpoint` not found")
        elif not kwds.get("api_version"):
            raise KeyError("Required `api_version` not found")
        elif kwds.get("timeout") and not isinstance(kwds.get("timeout"), Timeout):
            raise TypeError(
                "Invalid 'timeout' type detected:",
                type(kwds.get("timeout")),
                ", Please enter valid timeout using:\n'from httpx import Timeout'",
            )
        elif kwds.get("max_retries") and not isinstance(kwds.get("max_retries"), int):
            raise TypeError(
                "Invalid 'max_retries' type detected:,",
                type(kwds.get("max_retries")),
                ", Please enter a value that is 'int'",
            )
        else:
            pass

        # Initialize the Azure OpenAI client with provided parameters
        from openai import AzureOpenAI
        import tiktoken

        self.client = AzureOpenAI(
            azure_endpoint=kwds.get("azure_endpoint"),
            api_key=kwds.get("api_key"),
            api_version=kwds.get("api_version"),
            timeout=kwds.get("timeout")
            or Timeout(60.0, read=5.0, write=10.0, connect=2.0),
            max_retries=kwds.get("max_retries") or 2,
        )
        self.tiktoken = tiktoken

    def __call__(
        self,
        model_name: str,
        user_message: str,
        chat_hist: Optional[List[str]] = None,
        sampling_paras: Optional[Dict] = None,
        tools: Optional[List[Dict]] = None,
        tool_choice: Optional[str] = None,
        system_prompt: Optional[str] = None,
        do_json: Optional[bool] = False,
    ) -> Dict:
        """
        Generate a response from the specified model.

        Args:
            - model_name (str): Name of the Azure OpenAI model to use
            - user_message (Union[str, List[Dict[str, str]]]): The user's message or formatted messages
            - system_prompt (Optional[str], optional): System prompt to provide context to the model
            - chat_hist (Optional[List[str]], optional): Previous conversation history
            - sampling_paras (Optional[Dict], optional): Parameters for controlling the model's generation
            - tools (List[Dict], optional): List of tools to be used by the model.
            - tool_choice (Optional[str], optional): Specifies if and which tool the model must call — "none" for no tools, "auto" for automatic, "required" for mandatory, or a specific tool's name.
            - do_json (bool, optional): Whether to format the response as JSON. Defaults to False

        Returns:
            Dict: Response from the model or error information
        """
        try:
            # Format the chat history and user message
            messages = self._chat_formatter(
                user_message=user_message,
                chat_hist=chat_hist,
                system_prompt=system_prompt,
            )

            # Return early if message formatting failed
            if isinstance(messages, Dict):
                return messages

            # Default empty dictionaries
            sampling_paras = sampling_paras or {}

            # Check if tools and tool_choice are provided and format them
            tools = (
                [{"type": "function", "function": tool} for tool in tools]
                if tools
                else []
            )
            if tool_choice and tool_choice not in ["none", "auto", "required"]:
                if tool_choice in [
                    tool.get("function", {}).get("name") for tool in tools
                ]:
                    tool_choice = {
                        "type": "function",
                        "function": {"name": tool_choice},
                    }
                else:
                    return {
                        "error": 400,
                        "reason": f"Invalid tool_choice '{tool_choice}' provided. It must be one of ['none', 'auto', 'required'] or match a tool name in the provided tools.",
                    }

            # Call the OpenAI API with the formatted messages
            completion = self.client.chat.completions.create(
                model=model_name,
                messages=messages,
                tools=tools,
                tool_choice=tool_choice,
                response_format={"type": "json_object"}
                if do_json
                else {"type": "text"},
                **sampling_paras,
            )

            result = {
                "response": completion.choices[0].message.content or "",
                "metadata": {"usage": completion.usage.to_dict()},
            }

            if tools:
                tool_calls = (
                    [
                        tool_call.to_dict()
                        for tool_call in completion.choices[0].message.tool_calls
                    ]
                    if completion.choices[0].message.tool_calls
                    else []
                )
                for tool_call in tool_calls:
                    tool_call["function"]["arguments"] = (
                        json.loads(tool_call["function"]["arguments"])
                        if tool_call["function"]["arguments"]
                        else {}
                    )
                result["tools"] = tool_calls

            return result

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}

    def streaming(
        self,
        model_name: str,
        user_message: str,
        chat_hist: Optional[List[str]] = None,
        sampling_paras: Optional[Dict] = None,
        tools: Optional[List[Dict]] = None,
        tool_choice: Optional[str] = None,
        system_prompt: Optional[str] = None,
        do_json: Optional[bool] = False,
    ) -> Generator:
        """
        Stream responses from the specified model.

        Args:
            - model_name (str): Name of the Azure OpenAI model to use
            - user_message (Union[str, List[Dict[str, str]]]): The user's message or formatted messages
            - system_prompt (Optional[str], optional): System prompt to provide context to the model
            - chat_hist (Optional[List[str]], optional): Previous conversation history
            - sampling_paras (Optional[Dict], optional): Parameters for controlling the model's generation
            - tools (List[Dict], optional): List of tools to be used by the model.
            - tool_choice (Optional[str], optional): Specifies if and which tool the model must call — "none" for no tools, "auto" for automatic, "required" for mandatory, or a specific tool's name.
            - do_json (bool, optional): Whether to format the response as JSON. Defaults to False

        Yields:
            Generator: Chunks of the model's response or error information
        """
        try:
            # Format the chat history and user message
            messages = self._chat_formatter(
                user_message=user_message,
                chat_hist=chat_hist,
                system_prompt=system_prompt,
            )

            # Yield error and return early if message formatting failed
            if isinstance(messages, Dict):
                yield messages

            else:
                # Default empty dictionaries
                sampling_paras = sampling_paras or {}

                # Check if tools and tool_choice are provided and format them
                tools = (
                    [{"type": "function", "function": tool} for tool in tools]
                    if tools
                    else []
                )
                if tool_choice and tool_choice not in ["none", "auto", "required"]:
                    if tool_choice in [
                        tool.get("function", {}).get("name") for tool in tools
                    ]:
                        tool_choice = {
                            "type": "function",
                            "function": {"name": tool_choice},
                        }
                    else:
                        raise ValueError(
                            f"Invalid tool_choice '{tool_choice}' provided. It must be one of ['none', 'auto', 'required'] or match a tool name in the provided tools."
                        )

                # Start the streaming session
                completion = self.client.chat.completions.create(
                    model=model_name,
                    messages=messages,
                    stream=True,
                    tools=tools,
                    tool_choice=tool_choice,
                    response_format={"type": "json_object"}
                    if do_json
                    else {"type": "text"},
                    stream_options={"include_usage": True},
                    **sampling_paras,
                )

                response = ""
                usage = {}
                tool_calls = []
                tool_arg_buffer = ""

                # Stream text chunks as they become available
                for chunk in completion:
                    delta = chunk.choices[0].delta if chunk.choices else None
                    if delta and delta.content:
                        response += delta.content
                        yield delta.content
                    elif delta and delta.tool_calls:
                        tc = delta.tool_calls[0]
                        if tc.id:
                            tool_calls.append(tc.to_dict())
                        if tc.function and tc.function.arguments:
                            tool_arg_buffer += tc.function.arguments
                            try:
                                tool_calls[-1]["function"]["arguments"] = json.loads(
                                    tool_arg_buffer
                                )
                                tool_arg_buffer = ""
                            except json.JSONDecodeError:
                                continue
                    elif chunk.usage:
                        usage = chunk.usage.to_dict()

                # Format the final response with metadata
                result = {
                    "response": response,
                    "metadata": {"usage": usage},
                }

                if tools:
                    result["tools"] = tool_calls

                yield result
        except Exception as e:
            error_explainer(e)
            yield {"error": 500, "reason": str(e)}

    def _chat_formatter(
        self,
        user_message: str,
        chat_hist: Optional[List[str]] = None,
        system_prompt: Optional[str] = None,
    ) -> List[Dict]:
        """
        Format user messages and chat history for the Azure OpenAI API.

        Args:
            user_message (Union[str, List[Dict[str, str]]]): The user's message or formatted messages
            chat_hist (Optional[List[Dict[str, str]]], optional): Previous conversation history
            do_json (Optional[bool], optional): Whether to format the response as JSON. Defaults to False

        Returns:
            List[Dict]: Formatted messages in the structure expected by Azure OpenAI's API

        Raises:
            KeyError: If the user message format is invalid
        """
        try:
            messages = []

            # Add system prompt if provided
            if system_prompt:
                messages.append({"role": "system", "content": system_prompt})

            # Add chat history if provided
            if chat_hist:
                messages.extend(chat_hist)

            # Add user message
            messages.append({"role": "user", "content": user_message})

            return messages

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}

    def num_tokens_from_string(
        self, model_name: str = "gpt-3.5-turbo", string: str = None
    ) -> int:
        """Returns the number of tokens in a text string.

        Args:
        model_name (str): The tiktoken tokenizer for specifed model name
        string (str): String to calculate the tokens for

        Returns:
        int: Number of tokens"""
        if string and model_name:
            encoding = self.tiktoken.encoding_for_model(model_name)
            num_tokens = len(encoding.encode(string))
            return num_tokens
        else:
            return 0


class AsyncGenerate(object):
    """
    Asynchronous wrapper for Azure OpenAI's API client.

    This class provides methods for generating responses from Azure OpenAI's models
    both in streaming and non-streaming modes. It handles chat history formatting,
    error handling, and proper request configuration.
    """

    def __init__(self, **kwds) -> None:
        """
        Initialize Azure OpenAI client and set up API key.

        Args:
            - api_key (str): Azure OpenAI API key.
            - azure_endpoint (str): Azure OpenAI endpoint.
            - api_version (str): Azure OpenAI API version.
            - timeout (Timeout, optional): Request timeout parameter like connect, read, write. Default: 60.0, 5.0, 10.0, 2.0
            - max_retries (int, optional): Number of retries for the request. Default: 2

        Raises:
            - KeyError: If required parameters are not provided.
            - TypeError: If an invalid type is provided for a parameter
        """

        from httpx import Timeout

        # Validate input parameters
        if not kwds.get("api_key"):
            raise KeyError("Required `api_key` not found")
        elif not kwds.get("azure_endpoint"):
            raise KeyError("Required `azure_endpoint` not found")
        elif not kwds.get("api_version"):
            raise KeyError("Required `api_version` not found")
        elif kwds.get("timeout") and not isinstance(kwds.get("timeout"), Timeout):
            raise TypeError(
                "Invalid 'timeout' type detected:",
                type(kwds.get("timeout")),
                ", Please enter valid timeout using:\n'from httpx import Timeout'",
            )
        elif kwds.get("max_retries") and not isinstance(kwds.get("max_retries"), int):
            raise TypeError(
                "Invalid 'max_retries' type detected:,",
                type(kwds.get("max_retries")),
                ", Please enter a value that is 'int'",
            )
        else:
            pass

        # Initialize the Azure OpenAI client with provided parameters
        from openai import AsyncAzureOpenAI
        import tiktoken

        self.client = AsyncAzureOpenAI(
            azure_endpoint=kwds.get("azure_endpoint"),
            api_key=kwds.get("api_key"),
            api_version=kwds.get("api_version"),
            timeout=kwds.get("timeout")
            or Timeout(60.0, read=5.0, write=10.0, connect=2.0),
            max_retries=kwds.get("max_retries") or 2,
        )
        self.tiktoken = tiktoken

    async def __call__(
        self,
        model_name: str,
        user_message: str,
        request: Optional[Request] = None,
        chat_hist: Optional[List[str]] = None,
        sampling_paras: Optional[Dict] = None,
        tools: Optional[List[Dict]] = None,
        tool_choice: Optional[str] = None,
        system_prompt: Optional[str] = None,
        do_json: Optional[bool] = False,
    ) -> Dict:
        """
        Generate a response from the specified model.

        Args:
            - model_name (str): Name of the OpenAI model to use
            - user_message (Union[str, List[Dict[str, str]]]): The user's message or formatted messages
            - system_prompt (Optional[str], optional): System prompt to provide context to the model
            - request (Optional[Request], optional): FastAPI request object for connection tracking
            - chat_hist (Optional[List[str]], optional): Previous conversation history
            - sampling_paras (Optional[Dict], optional): Parameters for controlling the model's generation
            - tools (List[Dict], optional): List of tools to be used by the model.
            - tool_choice (Optional[str], optional): Specifies if and which tool the model must call — "none" for no tools, "auto" for automatic, "required" for mandatory, or a specific tool's name.
            - do_json (bool, optional): Whether to format the response as JSON. Defaults to False

        Returns:
            Dict: Response from the model or error information
        """
        try:
            # Format the chat history and user message
            messages = await self._chat_formatter(
                user_message=user_message,
                chat_hist=chat_hist,
                system_prompt=system_prompt,
            )

            # Return early if message formatting failed
            if isinstance(messages, Dict):
                return messages

            # Default empty dictionaries
            sampling_paras = sampling_paras or {}

            # Check if the request was disconnected
            if request and await request.is_disconnected():
                return {"error": 400, "reason": "request aborted by user"}

            # Check if tools and tool_choice are provided and format them
            tools = (
                [{"type": "function", "function": tool} for tool in tools]
                if tools
                else []
            )
            if tool_choice and tool_choice not in ["none", "auto", "required"]:
                if tool_choice in [
                    tool.get("function", {}).get("name") for tool in tools
                ]:
                    tool_choice = {
                        "type": "function",
                        "function": {"name": tool_choice},
                    }
                else:
                    return {
                        "error": 400,
                        "reason": f"Invalid tool_choice '{tool_choice}' provided. It must be one of ['none', 'auto', 'required'] or match a tool name in the provided tools.",
                    }

            # Call the Azure OpenAI API with the formatted messages
            completion = await self.client.chat.completions.create(
                model=model_name,
                messages=messages,
                tools=tools,
                tool_choice=tool_choice,
                response_format={"type": "json_object"}
                if do_json
                else {"type": "text"},
                **sampling_paras,
            )

            result = {
                "response": completion.choices[0].message.content or "",
                "metadata": {"usage": completion.usage.to_dict()},
            }

            if tools:
                tool_calls = (
                    [
                        tool_call.to_dict()
                        for tool_call in completion.choices[0].message.tool_calls
                    ]
                    if completion.choices[0].message.tool_calls
                    else []
                )
                for tool_call in tool_calls:
                    tool_call["function"]["arguments"] = (
                        json.loads(tool_call["function"]["arguments"])
                        if tool_call["function"]["arguments"]
                        else {}
                    )
                result["tools"] = tool_calls

            return result

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}

    async def streaming(
        self,
        model_name: str,
        user_message: str,
        request: Optional[Request] = None,
        chat_hist: Optional[List[str]] = None,
        sampling_paras: Optional[Dict] = None,
        tools: Optional[List[Dict]] = None,
        tool_choice: Optional[str] = None,
        system_prompt: Optional[str] = None,
        do_json: Optional[bool] = False,
    ) -> AsyncGenerator:
        """
        Stream responses from the specified model.

        Args:
            - model_name (str): Name of the OpenAI model to use
            - user_message (Union[str, List[Dict[str, str]]]): The user's message or formatted messages
            - system_prompt (Optional[str], optional): System prompt to provide context to the model
            - request (Optional[Request], optional): FastAPI request object for connection tracking
            - chat_hist (Optional[List[str]], optional): Previous conversation history
            - sampling_paras (Optional[Dict], optional): Parameters for controlling the model's generation
            - tools (List[Dict], optional): List of tools to be used by the model.
            - tool_choice (Optional[str], optional): Specifies if and which tool the model must call — "none" for no tools, "auto" for automatic, "required" for mandatory, or a specific tool's name.
            - do_json (bool, optional): Whether to format the response as JSON. Defaults to False

        Yields:
            AsyncGenerator: Chunks of the model's response or error information
        """
        try:
            # Format the chat history and user message
            messages = await self._chat_formatter(
                user_message=user_message,
                chat_hist=chat_hist,
                system_prompt=system_prompt,
            )

            # Yield error and return early if message formatting failed
            if isinstance(messages, Dict):
                yield messages

            else:
                # Default empty dictionaries
                sampling_paras = sampling_paras or {}

                # Check if tools and tool_choice are provided and format them
                tools = (
                    [{"type": "function", "function": tool} for tool in tools]
                    if tools
                    else []
                )
                if tool_choice and tool_choice not in ["none", "auto", "required"]:
                    if tool_choice in [
                        tool.get("function", {}).get("name") for tool in tools
                    ]:
                        tool_choice = {
                            "type": "function",
                            "function": {"name": tool_choice},
                        }
                    else:
                        raise ValueError(
                            f"Invalid tool_choice '{tool_choice}' provided. It must be one of ['none', 'auto', 'required'] or match a tool name in the provided tools."
                        )

                # Start the streaming session
                completion = await self.client.chat.completions.create(
                    model=model_name,
                    messages=messages,
                    stream=True,
                    tools=tools,
                    tool_choice=tool_choice,
                    response_format={"type": "json_object"}
                    if do_json
                    else {"type": "text"},
                    stream_options={"include_usage": True},
                    **sampling_paras,
                )

                response = ""
                usage = {}
                tool_calls = []
                tool_arg_buffer = ""

                # Stream text chunks as they become available
                async for chunk in completion:
                    if request and await request.is_disconnected():
                        yield {"error": 400, "reason": "request aborted by user"}
                        await completion.close()
                        break
                    else:
                        delta = chunk.choices[0].delta if chunk.choices else None
                        if delta and delta.content:
                            response += delta.content
                            yield delta.content
                        elif delta and delta.tool_calls:
                            tc = delta.tool_calls[0]
                            if tc.id:
                                tool_calls.append(tc.to_dict())
                            if tc.function and tc.function.arguments:
                                tool_arg_buffer += tc.function.arguments
                                try:
                                    tool_calls[-1]["function"]["arguments"] = (
                                        json.loads(tool_arg_buffer)
                                    )
                                    tool_arg_buffer = ""
                                except json.JSONDecodeError:
                                    continue
                        elif chunk.usage:
                            usage = chunk.usage.to_dict()

                # Format the final response with metadata
                result = {
                    "response": response,
                    "metadata": {"usage": usage},
                }

                if tools:
                    result["tools"] = tool_calls

                yield result
        except Exception as e:
            error_explainer(e)
            yield {"error": 500, "reason": str(e)}

    async def _chat_formatter(
        self,
        user_message: str,
        chat_hist: Optional[List[str]] = None,
        system_prompt: Optional[str] = None,
    ) -> List[Dict]:
        """
        Format user messages and chat history for the OpenAI API.

        Args:
            user_message (Union[str, List[Dict[str, str]]]): The user's message or formatted messages
            chat_hist (Optional[List[Dict[str, str]]], optional): Previous conversation history
            do_json (Optional[bool], optional): Whether to format the response as JSON. Defaults to False

        Returns:
            List[Dict]: Formatted messages in the structure expected by OpenAI's API

        Raises:
            KeyError: If the user message format is invalid
        """
        try:
            messages = []

            # Add system prompt if provided
            if system_prompt:
                messages.append({"role": "system", "content": system_prompt})

            # Add chat history if provided
            if chat_hist:
                messages.extend(chat_hist)

            # Add user message
            messages.append({"role": "user", "content": user_message})

            return messages

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}

    async def num_tokens_from_string(
        self, model_name: str = "gpt-3.5-turbo", string: str = None
    ) -> int:
        """Returns the number of tokens in a text string.

        Args:
        model_name (str): The tiktoken tokenizer for specifed model name
        string (str): String to calculate the tokens for

        Returns:
        int: Number of tokens"""
        if string and model_name:
            encoding = self.tiktoken.encoding_for_model(model_name)
            num_tokens = len(encoding.encode(string))
            return num_tokens
        else:
            return 0



from typing import (
    Any,
    Dict,
    List,
    Optional,
    Union,
    Generator,
    AsyncGenerator,
)
from fastapi import Request
from orichain import error_explainer


class Generate(object):
    """
    Synchronous wrapper for Google's API client for Vertex AI.

    This class provides methods for generating responses from Google's models
    both in streaming and non-streaming modes. It handles chat history formatting,
    error handling, and proper request configuration.
    """

    def __init__(self, **kwds) -> None:
        """
        Initialize Google client for Vertex AI and set up API key.

        Args:
            - api_key (str): Vertex AI API key
            - credentials (google.auth.credentials.Credentials): The credentials to use for authentication when calling the Vertex AI APIs.
            - project (str): The Google Cloud project ID to use for quota.
            - location (str): The location to send API requests to (for example, us-central1).
            - http_options (types.HttpOptions, optional): HTTP options to be used in each of the requests. Default is None
            - debug_config (DebugConfig, optional): Configuration options that change client network behavior when testing. Default is None

        Raises:
            - KeyError: If required parameters are not provided.
            - TypeError: If an invalid type is provided for a parameter
        """

        from google.genai import Client, types
        from google.auth.credentials import Credentials
        from google.genai.client import DebugConfig

        # Validate input parameters
        if not kwds.get("api_key"):
            raise KeyError("Required `api_key` not found")
        elif not kwds.get("project"):
            raise KeyError("Required `project` not found")
        elif not kwds.get("location"):
            raise KeyError("Required `location` not found")
        elif kwds.get("credentials") and not isinstance(
            kwds.get("credentials"), Credentials
        ):
            raise TypeError(
                "Invalid 'credentials' type detected:",
                type(kwds.get("credentials")),
                ", Please enter valid credentials using:\n'from google.auth.credentials import Credentials'",
            )
        elif kwds.get("http_options") and not isinstance(
            kwds.get("http_options"), types.HttpOptions
        ):
            raise TypeError(
                "Invalid 'http_options' type detected:",
                type(kwds.get("http_options")),
                ", Please enter valid http_options using:\n'from google.genai.types import HttpOptions'",
            )
        elif kwds.get("debug_config") and not isinstance(
            kwds.get("debug_config"), DebugConfig
        ):
            raise TypeError(
                "Invalid 'debug_config' type detected:",
                type(kwds.get("debug_config")),
                ", Please enter valid debug_config using:\n'from google.genai.client import DebugConfig'",
            )
        else:
            pass

        # Initialize the Google client with provided parameters
        self.client: Client = Client(
            vertexai=True,
            api_key=kwds.get("api_key"),
            credentials=kwds.get("credentials"),
            project=kwds.get("project"),
            location=kwds.get("location"),
            http_options=kwds.get("http_options", types.HttpOptions(timeout=7000)),
            debug_config=kwds.get("debug_config"),
        )

        self.fields = [
            "content",
            "video_metadata",
            "thought",
            "inline_data",
            "code_execution_result",
            "executable_code",
            "file_data",
            "function_call",
            "function_response",
        ]

        self.types = types

    def __call__(
        self,
        model_name: str,
        user_message: Union[str, List[Union[Dict[str, str], Any]]],
        chat_hist: Optional[List[str]] = None,
        sampling_paras: Optional[Dict] = None,
        tools: Optional[List[Dict]] = None,
        tool_choice: Optional[str] = None,
        system_prompt: Optional[str] = None,
        do_json: Optional[bool] = False,
        **kwds: Any,
    ) -> Dict:
        """
        Generate a response from the specified model.

        Args:
            - model_name (str): Name of the Vertex AI model to use
            - user_message (Union[str, List[Union[Dict[str, str], Any]]]): The user's message or formatted messages
            - system_prompt (Optional[str], optional): System prompt to provide context to the model
            - chat_hist (Optional[List[str]], optional): Previous conversation history
            - sampling_paras (Optional[Dict], optional): Parameters for controlling the model's generation by GenerateContentConfig
            - tools (List[Dict], optional): List of tools to be used by the model.
            - tool_choice (Optional[str], optional): Specifies if and which tool the model must call — "none" for no tools, "auto" for automatic, "required" for mandatory, or a specific tool's name.
            - do_json (bool, optional): Whether to format the response as JSON. Defaults to False
            - **kwds: Additional keyword arguments to pass to the client

        Returns:
            Dict: Response from the model or error information
        """
        try:
            # Format the chat history and user message
            messages = self._chat_formatter(
                chat_hist=chat_hist,
            )

            # Return early if message formatting failed
            if isinstance(messages, Dict):
                return messages

            # Default empty dictionaries
            sampling_paras = sampling_paras or {}

            # Checking tool_choice and formatting tool_config
            tool_config = None
            if tool_choice:
                if tool_choice == "required":
                    tool_choice = "any"
                if tool_choice in ["none", "auto", "any"]:
                    tool_config = self.types.ToolConfig(
                        function_calling_config=self.types.FunctionCallingConfig(
                            mode=tool_choice.upper()
                        )
                    )
                elif tool_choice in [tool.get("name") for tool in tools]:
                    tool_config = self.types.ToolConfig(
                        function_calling_config=self.types.FunctionCallingConfig(
                            mode="ANY", allowed_function_names=[tool_choice]
                        )
                    )
                else:
                    return {
                        "error": 400,
                        "reason": f"Invalid tool_choice '{tool_choice}' provided. It must be one of ['none', 'auto', 'required'] or match a tool name in the provided tools.",
                    }

            # Create new chat session with Google API with the formatted messages
            chat_session = self.client.chats.create(
                model=model_name,
                config=kwds.get(
                    "config",
                    self.types.GenerateContentConfig(
                        system_instruction=system_prompt,
                        response_mime_type=(
                            "application/json"
                            if do_json
                            else kwds.get("response_mime_type") or "text/plain"
                        ),
                        tools=[self.types.Tool(function_declarations=tools)]
                        if tools
                        else [],
                        tool_config=tool_config,
                        **sampling_paras,
                    ),
                ),
                history=messages,
            )

            response = chat_session.send_message(message=user_message)

            # Fetching responses from the LLM for tools and text
            result = {
                "response": response.text or "",
                "metadata": {"usage": response.usage_metadata.to_json_dict()},
            }

            if tools:
                tool_calls = []
                if response.function_calls:
                    for tool in response.function_calls:
                        tool_calls.append(
                            {
                                "id": tool.id,
                                "function": {"name": tool.name, "arguments": tool.args},
                            }
                        )
                result["tools"] = tool_calls

            return result

        except Exception as e:
            error_explainer(e=e)
            return {"error": 500, "reason": str(e)}

    def streaming(
        self,
        model_name: str,
        user_message: Union[str, List[Union[Dict[str, str], Any]]],
        chat_hist: Optional[List[str]] = None,
        sampling_paras: Optional[Dict] = None,
        tools: Optional[List[Dict]] = None,
        tool_choice: Optional[str] = None,
        system_prompt: Optional[str] = None,
        do_json: Optional[bool] = False,
        **kwds: Any,
    ) -> Generator:
        """
        Stream responses from the specified model.

        Args:
            - model_name (str): Name of the Google model to use
            - user_message (Union[str, List[Union[Dict[str, str], Any]]]): The user's message or formatted messages
            - system_prompt (Optional[str], optional): System prompt to provide context to the model
            - chat_hist (Optional[List[str]], optional): Previous conversation history
            - sampling_paras (Optional[Dict], optional): Parameters for controlling the model's generation by GenerateContentConfig
            - tools (List[Dict], optional): List of tools to be used by the model.
            - tool_choice (Optional[str], optional): Specifies if and which tool the model must call — "none" for no tools, "auto" for automatic, "required" for mandatory, or a specific tool's name.
            - do_json (bool, optional): Whether to format the response as JSON. Defaults to False
            - **kwds: Additional keyword arguments to pass to the client

        Yields:
            AsyncGenerator: Chunks of the model's response or error information
        """
        try:
            # Format the chat history and user message
            messages = self._chat_formatter(
                chat_hist=chat_hist,
            )

            # Yield error and return early if message formatting failed
            if isinstance(messages, Dict):
                yield messages

            else:
                # Default empty dictionaries
                sampling_paras = sampling_paras or {}

                # Checking tool_choice and formatting tool_config
                tool_config = None
                if tool_choice:
                    if tool_choice == "required":
                        tool_choice = "any"
                    if tool_choice in ["none", "auto", "any"]:
                        tool_config = self.types.ToolConfig(
                            function_calling_config=self.types.FunctionCallingConfig(
                                mode=tool_choice.upper()
                            )
                        )
                    elif tool_choice in [tool.get("name") for tool in tools]:
                        tool_config = self.types.ToolConfig(
                            function_calling_config=self.types.FunctionCallingConfig(
                                mode="ANY", allowed_function_names=[tool_choice]
                            )
                        )
                    else:
                        raise ValueError(
                            f"Invalid tool_choice '{tool_choice}' provided. It must be one of ['none', 'auto', 'required'] or match a tool name in the provided tools."
                        )

                # Create new chat session with Google API with the formatted messages
                chat_session = self.client.chats.create(
                    model=model_name,
                    config=kwds.get(
                        "config",
                        self.types.GenerateContentConfig(
                            system_instruction=system_prompt,
                            response_mime_type=(
                                "application/json"
                                if do_json
                                else kwds.get("response_mime_type") or "text/plain"
                            ),
                            tools=[self.types.Tool(function_declarations=tools)]
                            if tools
                            else [],
                            tool_config=tool_config,
                            **sampling_paras,
                        ),
                    ),
                    history=messages,
                )

                result: Dict = {"response": ""}
                tool_calls = []

                for chunk in chat_session.send_message_stream(message=user_message):
                    if chunk.text:
                        result["response"] = result["response"] + chunk.text
                        yield chunk.text
                    elif chunk.function_calls:
                        for tool in chunk.function_calls:
                            tool_calls.append(
                                {
                                    "id": tool.id,
                                    "function": {
                                        "name": tool.name,
                                        "arguments": tool.args,
                                    },
                                }
                            )
                    if chunk.usage_metadata:
                        result.update({"usage": chunk.usage_metadata.to_json_dict()})

                if tools:
                    result["tools"] = tool_calls

                yield result
        except Exception as e:
            error_explainer(e)
            yield {"error": 500, "reason": str(e)}

    def _chat_formatter(
        self,
        chat_hist: Optional[List[Union[Dict[str, Union[str, Dict, Any]], Any]]] = None,
    ) -> Union[List[Union[Dict, Any]], Dict, None]:
        """
        Format user messages and chat history for the Google API.

        Args:
            chat_hist (Optional[List[Union[Dict[str, Union[str, Dict, Any]], types.Content]]], optional): Previous conversation history

        Returns:
            List[Dict]: Formatted messages in the structure expected by Google's API
        """
        try:
            # Add chat history if provided
            if chat_hist:
                messages = []
                parts = []
                for chat in chat_hist:
                    if isinstance(chat, Dict):
                        if chat.get("role") == "tool":
                            messages.append(chat)
                        else:
                            for field in self.fields:
                                message = self.types.Part()
                                if field in chat:
                                    setattr(
                                        message,
                                        field if field != "content" else "text",
                                        chat[field],
                                    )
                                parts.append(message)

                            content = self.types.Content(
                                role="user" if chat.get("role") == "user" else "model",
                                parts=parts,
                            )
                            messages.append(content)
                    else:
                        messages.append(chat)
                return messages
            else:
                None

        except Exception as e:
            error_explainer(e=e)
            return {"error": 500, "reason": str(e)}


class AsyncGenerate(object):
    """
    Asynchronous wrapper for Google's API client for Vertex AI.

    This class provides methods for generating responses from Google's models
    both in streaming and non-streaming modes. It handles chat history formatting,
    error handling, and proper request configuration.
    """

    def __init__(self, **kwds) -> None:
        """
        Initialize Google client for Vertex AI and set up API key.

        Args:
            - api_key (str): Vertex AI API key
            - credentials (google.auth.credentials.Credentials): The credentials to use for authentication when calling the Vertex AI APIs.
            - project (str): The Google Cloud project ID to use for quota.
            - location (str): The location to send API requests to (for example, us-central1).
            - http_options (types.HttpOptions, optional): HTTP options to be used in each of the requests. Default is None
            - debug_config (DebugConfig, optional): Configuration options that change client network behavior when testing. Default is None

        Raises:
            - KeyError: If required parameters are not provided.
            - TypeError: If an invalid type is provided for a parameter
        """

        from google.genai import Client, types
        from google.auth.credentials import Credentials
        from google.genai.client import DebugConfig

        # Validate input parameters
        if not kwds.get("api_key"):
            raise KeyError("Required `api_key` not found")
        elif not kwds.get("project"):
            raise KeyError("Required `project` not found")
        elif not kwds.get("location"):
            raise KeyError("Required `location` not found")
        elif kwds.get("credentials") and not isinstance(
            kwds.get("credentials"), Credentials
        ):
            raise TypeError(
                "Invalid 'credentials' type detected:",
                type(kwds.get("credentials")),
                ", Please enter valid credentials using:\n'from google.auth.credentials import Credentials'",
            )
        elif kwds.get("http_options") and not isinstance(
            kwds.get("http_options"), types.HttpOptions
        ):
            raise TypeError(
                "Invalid 'http_options' type detected:",
                type(kwds.get("http_options")),
                ", Please enter valid http_options using:\n'from google.genai.types import HttpOptions'",
            )
        elif kwds.get("debug_config") and not isinstance(
            kwds.get("debug_config"), DebugConfig
        ):
            raise TypeError(
                "Invalid 'debug_config' type detected:",
                type(kwds.get("debug_config")),
                ", Please enter valid debug_config using:\n'from google.genai.client import DebugConfig'",
            )
        else:
            pass

        # Initialize the Google client with provided parameters
        self.client: Client = Client(
            vertexai=True,
            api_key=kwds.get("api_key"),
            credentials=kwds.get("credentials"),
            project=kwds.get("project"),
            location=kwds.get("location"),
            http_options=kwds.get("http_options", types.HttpOptions(timeout=7000)),
            debug_config=kwds.get("debug_config"),
        )

        self.fields = [
            "content",
            "video_metadata",
            "thought",
            "inline_data",
            "code_execution_result",
            "executable_code",
            "file_data",
            "function_call",
            "function_response",
        ]

        self.types = types

    async def __call__(
        self,
        model_name: str,
        user_message: Union[str, List[Union[Dict[str, str], Any]]],
        request: Optional[Request] = None,
        chat_hist: Optional[List[str]] = None,
        sampling_paras: Optional[Dict] = None,
        tools: Optional[List[Dict]] = None,
        tool_choice: Optional[str] = None,
        system_prompt: Optional[str] = None,
        do_json: Optional[bool] = False,
        **kwds: Any,
    ) -> Dict:
        """
        Generate a response from the specified model.

        Args:
            - model_name (str): Name of the Vertex AI model to use
            - user_message (Union[str, List[Union[Dict[str, str], Any]]]): The user's message or formatted messages
            - system_prompt (Optional[str], optional): System prompt to provide context to the model
            - request (Optional[Request], optional): FastAPI request object for connection tracking
            - chat_hist (Optional[List[str]], optional): Previous conversation history
            - sampling_paras (Optional[Dict], optional): Parameters for controlling the model's generation by GenerateContentConfig
            - tools (List[Dict], optional): List of tools to be used by the model.
            - tool_choice (Optional[str], optional): Specifies if and which tool the model must call — "none" for no tools, "auto" for automatic, "required" for mandatory, or a specific tool's name.
            - do_json (bool, optional): Whether to format the response as JSON. Defaults to False
            - **kwds: Additional keyword arguments to pass to the client

        Returns:
            Dict: Response from the model or error information
        """
        try:
            # Format the chat history and user message
            messages = await self._chat_formatter(
                chat_hist=chat_hist,
            )

            # Return early if message formatting failed
            if isinstance(messages, Dict):
                return messages

            # Default empty dictionaries
            sampling_paras = sampling_paras or {}

            # Checking tool_choice and formatting tool_config
            tool_config = None
            if tool_choice:
                if tool_choice == "required":
                    tool_choice = "any"
                if tool_choice in ["none", "auto", "any"]:
                    tool_config = self.types.ToolConfig(
                        function_calling_config=self.types.FunctionCallingConfig(
                            mode=tool_choice.upper()
                        )
                    )
                elif tool_choice in [tool.get("name") for tool in tools]:
                    tool_config = self.types.ToolConfig(
                        function_calling_config=self.types.FunctionCallingConfig(
                            mode="ANY", allowed_function_names=[tool_choice]
                        )
                    )
                else:
                    return {
                        "error": 400,
                        "reason": f"Invalid tool_choice '{tool_choice}' provided. It must be one of ['none', 'auto', 'required'] or match a tool name in the provided tools.",
                    }

            # Check if the request was disconnected
            if request and await request.is_disconnected():
                return {"error": 400, "reason": "request aborted by user"}

            # Create new chat session with Google API with the formatted messages
            chat_session = self.client.aio.chats.create(
                model=model_name,
                config=kwds.get(
                    "config",
                    self.types.GenerateContentConfig(
                        system_instruction=system_prompt,
                        response_mime_type=(
                            "application/json"
                            if do_json
                            else kwds.get("response_mime_type") or "text/plain"
                        ),
                        tools=[self.types.Tool(function_declarations=tools)]
                        if tools
                        else [],
                        tool_config=tool_config,
                        **sampling_paras,
                    ),
                ),
                history=messages,
            )

            response = await chat_session.send_message(message=user_message)

            # Fetching responses from the LLM for tools and text
            result = {
                "response": response.text or "",
                "metadata": {"usage": response.usage_metadata.to_json_dict()},
            }

            if tools:
                tool_calls = []
                if response.function_calls:
                    for tool in response.function_calls:
                        tool_calls.append(
                            {
                                "id": tool.id,
                                "function": {"name": tool.name, "arguments": tool.args},
                            }
                        )
                result["tools"] = tool_calls

            return result

        except Exception as e:
            error_explainer(e=e)
            return {"error": 500, "reason": str(e)}

    async def streaming(
        self,
        model_name: str,
        user_message: Union[str, List[Union[Dict[str, str], Any]]],
        request: Optional[Request] = None,
        chat_hist: Optional[List[str]] = None,
        sampling_paras: Optional[Dict] = None,
        tools: Optional[List[Dict]] = None,
        tool_choice: Optional[str] = None,
        system_prompt: Optional[str] = None,
        do_json: Optional[bool] = False,
        **kwds: Any,
    ) -> AsyncGenerator:
        """
        Stream responses from the specified model.

        Args:
            - model_name (str): Name of the Google model to use
            - user_message (Union[str, List[Union[Dict[str, str], Any]]]): The user's message or formatted messages
            - system_prompt (Optional[str], optional): System prompt to provide context to the model
            - request (Optional[Request], optional): FastAPI request object for connection tracking
            - chat_hist (Optional[List[str]], optional): Previous conversation history
            - sampling_paras (Optional[Dict], optional): Parameters for controlling the model's generation by GenerateContentConfig
            - tools (List[Dict], optional): List of tools to be used by the model.
            - tool_choice (Optional[str], optional): Specifies if and which tool the model must call — "none" for no tools, "auto" for automatic, "required" for mandatory, or a specific tool's name.
            - do_json (bool, optional): Whether to format the response as JSON. Defaults to False
            - **kwds: Additional keyword arguments to pass to the client

        Yields:
            AsyncGenerator: Chunks of the model's response or error information
        """
        try:
            # Format the chat history and user message
            messages = await self._chat_formatter(
                chat_hist=chat_hist,
            )

            # Yield error and return early if message formatting failed
            if isinstance(messages, Dict):
                yield messages

            else:
                # Default empty dictionaries
                sampling_paras = sampling_paras or {}

                # Checking tool_choice and formatting tool_config
                tool_config = None
                if tool_choice:
                    if tool_choice == "required":
                        tool_choice = "any"
                    if tool_choice in ["none", "auto", "any"]:
                        tool_config = self.types.ToolConfig(
                            function_calling_config=self.types.FunctionCallingConfig(
                                mode=tool_choice.upper()
                            )
                        )
                    elif tool_choice in [tool.get("name") for tool in tools]:
                        tool_config = self.types.ToolConfig(
                            function_calling_config=self.types.FunctionCallingConfig(
                                mode="ANY", allowed_function_names=[tool_choice]
                            )
                        )
                    else:
                        raise ValueError(
                            f"Invalid tool_choice '{tool_choice}' provided. It must be one of ['none', 'auto', 'required'] or match a tool name in the provided tools."
                        )

                # Create new chat session with Google API with the formatted messages
                chat_session = self.client.aio.chats.create(
                    model=model_name,
                    config=kwds.get(
                        "config",
                        self.types.GenerateContentConfig(
                            system_instruction=system_prompt,
                            response_mime_type=(
                                "application/json"
                                if do_json
                                else kwds.get("response_mime_type") or "text/plain"
                            ),
                            tools=[self.types.Tool(function_declarations=tools)]
                            if tools
                            else [],
                            tool_config=tool_config,
                            **sampling_paras,
                        ),
                    ),
                    history=messages,
                )

                result: Dict = {"response": ""}
                tool_calls = []

                async for chunk in await chat_session.send_message_stream(
                    message=user_message
                ):
                    if request and await request.is_disconnected():
                        yield {"error": 400, "reason": "request aborted by user"}
                        break

                    if chunk.text:
                        result["response"] = result["response"] + chunk.text
                        yield chunk.text
                    elif chunk.function_calls:
                        for tool in chunk.function_calls:
                            tool_calls.append(
                                {
                                    "id": tool.id,
                                    "function": {
                                        "name": tool.name,
                                        "arguments": tool.args,
                                    },
                                }
                            )
                    if chunk.usage_metadata:
                        result.update({"usage": chunk.usage_metadata.to_json_dict()})

                if tools:
                    result["tools"] = tool_calls

                yield result
        except Exception as e:
            error_explainer(e)
            yield {"error": 500, "reason": str(e)}

    async def _chat_formatter(
        self,
        chat_hist: Optional[List[Union[Dict[str, Union[str, Dict, Any]], Any]]] = None,
    ) -> Union[List[Union[Dict, Any]], Dict, None]:
        """
        Format user messages and chat history for the Google API.

        Args:
            chat_hist (Optional[List[Union[Dict[str, Union[str, Dict, Any]], types.Content]]], optional): Previous conversation history

        Returns:
            List[Dict]: Formatted messages in the structure expected by Google's API
        """
        try:
            # Add chat history if provided
            if chat_hist:
                messages = []
                parts = []
                for chat in chat_hist:
                    if isinstance(chat, Dict):
                        if chat.get("role") == "tool":
                            messages.append(chat)
                        else:
                            for field in self.fields:
                                message = self.types.Part()
                                if field in chat:
                                    setattr(
                                        message,
                                        field if field != "content" else "text",
                                        chat[field],
                                    )
                                parts.append(message)

                            content = self.types.Content(
                                role="user" if chat.get("role") == "user" else "model",
                                parts=parts,
                            )
                            messages.append(content)
                    else:
                        messages.append(chat)
                return messages
            else:
                None

        except Exception as e:
            error_explainer(e=e)
            return {"error": 500, "reason": str(e)}



from typing import Any, List, Dict, Union
from orichain.embeddings import (
    openai_embeddings,
    awsbedrock_embeddings,
    stransformers_embeddings,
    azureopenai_embeddings,
    gcp_gemini_embeddings,
    gcp_vertex_embeddings,
    togetherai_embeddings,
)
import warnings
from orichain import hf_repo_exists

DEFUALT_EMBEDDING_MODEL = "text-embedding-3-small"
DEFAULT_MODEL_PROVIDER = "OpenAI"
SUPPORTED_MODELS = {
    "OpenAI": [
        "text-embedding-ada-002",
        "text-embedding-3-large",
        "text-embedding-3-small",
    ],
    "AzureOpenAI": [
        "text-embedding-ada-002",
        "text-embedding-3-large",
        "text-embedding-3-small",
    ],
    "AWSBedrock": [
        "amazon.titan-embed-text-v1",
        "amazon.titan-embed-text-v2:0",
        "cohere.embed-english-v3",
        "cohere.embed-multilingual-v3",
    ],
    "GoogleGemini": [
        "text-embedding-004",
        "gemini-embedding-exp-03-07",
        "embedding-001",
    ],
    "GoogleVertexAI": [
        "text-multilingual-embedding-002",
        "text-embedding-004",
        "text-embedding-005",
        "gemini-embedding-001",
        "gemini-embedding-exp-03-07",
        "embedding-001",
    ],
    "TogetherAI": [
        "togethercomputer/m2-bert-80M-32k-retrieval",
        "BAAI/bge-large-en-v1.5",
        "BAAI/bge-base-en-v1.5",
        "Alibaba-NLP/gte-modernbert-base",
        "intfloat/multilingual-e5-large-instruct",
    ],
}


class EmbeddingModel(object):
    """Synchronus Base class for embedding generation.

    This class provides a unified interface to interact with different embedding models from providers such as OpenAI, AWS Bedrock, Google Gemini and Vertex AI, Azure OpenAI, and SentenceTransformers."""

    default_model = DEFUALT_EMBEDDING_MODEL
    default_model_provider = DEFAULT_MODEL_PROVIDER
    supported_models = SUPPORTED_MODELS
    model_handler = {
        "OpenAI": openai_embeddings.Embed,
        "AWSBedrock": awsbedrock_embeddings.Embed,
        "SentenceTransformers": stransformers_embeddings.Embed,
        "AzureOpenAI": azureopenai_embeddings.Embed,
        "GoogleGemini": gcp_gemini_embeddings.Embed,
        "GoogleVertexAI": gcp_vertex_embeddings.Embed,
        "TogetherAI": togetherai_embeddings.Embed,
    }

    def __init__(self, **kwds: Any) -> None:
        """Initialize the Embedding Models class with the required parameters.

        Args:
            - model_name (str, optional): Name of the model to be used. Default: "text-embedding-3-small"
            - provider (str, optional): Name of the model provider. Default: "OpenAI". Allowed values:
                - OpenAI
                - AWSBedrock
                - GoogleGemini
                - GoogleVertexAI
                - AzureOpenAI
                - TogetherAI
                - SentenceTransformers

            **Authentication Arguments by provider:**

                **OpenAI models:**
                    - api_key (str): OpenAI API key.
                    - timeout (Timeout, optional): Request timeout parameter like connect, read, write. Default: 60.0, 5.0, 10.0, 2.0
                    - max_retries (int, optional): Number of retries for the request. Default: 2

                **AWS Bedrock models:**
                    - aws_access_key (str): AWS access key.
                    - aws_secret_key (str): AWS secret key.
                    - aws_region (str): AWS region name.
                    - config (Config, optional):
                        - connect_timeout (float or int, optional): The time in seconds till a timeout exception is thrown when attempting to make a connection. Default: 60
                        - read_timeout: (float or int, optional): The time in seconds till a timeout exception is thrown when attempting to read from a connection. Default: 60
                        - region_name (str, optional): region name Note: If specifing config you need to still pass region_name even if you have already passed in aws_region
                        - max_pool_connections: The maximum number of connections to keep in a connection pool. Defualt: 10
                        - retries (Dict, optional):
                            - total_max_attempts: Number of retries for the request. Default: 2

                **Google Gemini models:**
                    - api_key (str): Gemini API key
                    - http_options (types.HttpOptions, optional): HTTP options to be used in each of the requests. Default is None
                    - debug_config (DebugConfig, optional): Configuration options that change client network behavior when testing. Default is None

                **Google Vertex AI models:**
                    - api_key (str): Vertex AI API key
                    - credentials (google.auth.credentials.Credentials): The credentials to use for authentication when calling the Vertex AI APIs.
                    - project (str): The Google Cloud project ID to use for quota.
                    - location (str): The location to send API requests to (for example, us-central1).
                    - http_options (types.HttpOptions, optional): HTTP options to be used in each of the requests. Default is None
                    - debug_config (DebugConfig, optional): Configuration options that change client network behavior when testing. Default is None

                **Sentence Transformers models:**
                    - model_download_path (str, optional): Path to download the model. Default: "/home/ubuntu/projects/models/embedding_models"
                    - device (str, optional): Device to run the model. Default: "cpu"
                    - trust_remote_code (bool, optional): Trust remote code. Default: False
                    - token (str, optional): Hugging Face API token

                **Azure OpenAI models:**
                    - api_key (str): Azure OpenAI API key.
                    - azure_endpoint (str): Azure OpenAI endpoint.
                    - api_version (str): Azure OpenAI API version.
                    - timeout (Timeout, optional): Request timeout parameter like connect, read, write. Default: 60.0, 5.0, 10.0, 2.0
                    - max_retries (int, optional): Number of retries for the request. Default: 2

                **TogetherAI models:**
                    - api_key (str): TogetherAI API key.
                    - timeout (Timeout, optional): Request timeout in seconds. Default: 60
                    - max_retries (int, optional): Number of retries for the request. Default: 2

        Raises:
            - ValueError: If the model is not supported
            - KeyError: If required parameters are missing
            - TypeError: If the type of the parameter is incorrect
            - ImportError: If the required library is not installed

        Warns:
            - UserWarning: If no model_name is provided, defaulting to `text-embedding-3-small`
        """

        # Check if the model name is provided
        if not kwds.get("model_name"):
            warnings.warn(
                f"\nNo 'model_name' specified, hence defaulting to {self.default_model}",
                UserWarning,
            )
        if not kwds.get("provider"):
            warnings.warn(
                f"\nNo 'provider' specified, hence defaulting to {self.default_model_provider}",
                UserWarning,
            )

        self.model_name = kwds.get("model_name", self.default_model)
        self.model_provider = kwds.get("provider", self.default_model_provider)

        # Validating model name and model provider name
        if self.model_provider not in self.model_handler:
            raise ValueError(
                f"\nUnsupported model provider: {self.model_provider}\nSupported providers are:"
                f"\n- " + "\n- ".join(list(self.model_handler.keys()))
            )
        elif self.model_provider == "SentenceTransformers":
            repo_check = hf_repo_exists(
                repo_id=self.model_name,
                repo_type=kwds.get("repo_type"),
                token=kwds.get("token"),
            )
            if not repo_check:
                raise ValueError(
                    f"\nThe Huggingface repository '{self.model_name}' does not exist. \nPlease ensure you provide the full repository path in 'model_name'."
                )
        elif self.model_name not in self.supported_models.get(self.model_provider):
            warnings.warn(
                f"\nModel {self.model_name} for provider {self.model_provider} is not supported by Orichain. Supported models for {self.model_provider} are: [{', '.join(self.supported_models.get(self.model_provider))}] \nPlease make sure you're using the correct 'model_name' and 'provider'",
                UserWarning,
            )
        else:
            pass

        # Initialize the model
        self.model = self.model_handler.get(self.model_provider)(**kwds)

    def __call__(
        self, user_message: Union[str, List[str]], **kwds: Any
    ) -> Union[List[float], List[List[float]], Dict]:
        """Get embeddings for the given text(s).

        Args:
            - user_message (Union[str, List[str]]): Input text or list of texts

            **Generation Arguments by provider:**

                **OpenAI & Azure OpenAI models:**
                    - model_name (str, optional): Name of the embedding model to use.

                **Google Gemini & Vertex AI models:**
                    - model_name (str, optional): Name of the embedding model to use
                    - config (google.genai.types.EmbedContentConfig, optional): Optional model configuration parameters provided to the client.models.embed_content API.

                **AWS Bedrock models:**
                    - model_name (str, optional): Name of the embedding model to use.

                    **Cohere Embedding Models:**
                        - input_type (Literal["search_query", "search_document", "classification", "clustering", "image"], optional): Type of input text. Default: "search_query"
                        - embedding_types (str, optional): Specifies the types of embeddings you want to have returned. Can be one or more of the following types: 'float', 'int8', 'uint8', 'binary', 'ubinary'
                        - truncate (Literal["NONE", "START", "END"], optional): Specifies how the API handles inputs longer than the maximum token length. Use one of the following:
                            - NONE – (Default) Returns an error when the input exceeds the maximum input token length.
                            - START – Discards the start of the input.
                            - END – Discards the end of the input.

                    **Amazon Titan Embeddings G1 Models:**
                        - dimensions (int, optional): Output dimensions. Default: 1024 (Output dimensions can be: 256, 512 and 1024)
                        - normalize (bool, optional): Normalize the output. Default: True (As recommended in docs for RAG)

                **Sentence Transformers models:**
                    - prompt_name (str, optional): The name of the prompt to use for encoding. Must be a key in the `prompts` dictionary,
                      which is either set in the constructor or loaded from the model configuration. For example if
                      ``prompt_name`` is "query" and the ``prompts`` is {"query": "query: ", ...}, then the sentence "What
                      is the capital of France?" will be encoded as "query: What is the capital of France?" because the sentence
                      is appended to the prompt. If ``prompt`` is also set, this argument is ignored. Defaults to None.
                    - prompt (str, optional): The prompt to use for encoding. For example, if the prompt is "query: ", then the
                      sentence "What is the capital of France?" will be encoded as "query: What is the capital of France?"
                      because the sentence is appended to the prompt. If ``prompt`` is set, ``prompt_name`` is ignored. Defaults to None.
                    - output_value (Literal["sentence_embedding", "token_embeddings"], optional): The type of embeddings to return:
                      "sentence_embedding" to get sentence embeddings, "token_embeddings" to get wordpiece token embeddings, and `None`,
                      to get all output values. Defaults to "sentence_embedding".
                    - show_progress_bar (bool, optional): Whether to output a progress bar when encode sentences. Defaults to False.
                    - precision (Literal["float32", "int8", "uint8", "binary", "ubinary"], optional): The precision to use for the embeddings.
                      Can be "float32", "int8", "uint8", "binary", or "ubinary". All non-float32 precisions are quantized embeddings.
                      Quantized embeddings are smaller in size and faster to compute, but may have a lower accuracy. They are useful for
                      reducing the size of the embeddings of a corpus for semantic search, among other tasks. Defaults to "float32".
                    - batch_size (int, optional): The batch size used for the computation. Defaults to 32.
                    - convert_to_numpy (bool, optional): Whether the output should be a list of numpy vectors. If False, it is a list of PyTorch tensors.
                      Defaults to False.
                    - convert_to_tensor (bool, optional): Whether the output should be one large tensor. Overwrites `convert_to_numpy`.
                      Defaults to False.
                    - device (str, optional): Which :class:`torch.device` to use for the computation. Defaults to None.
                    - normalize_embeddings (bool, optional): Whether to normalize returned vectors to have length 1. In that case,
                      the faster dot-product (util.dot_score) instead of cosine similarity can be used. Defaults to False.

        Returns:
            (Union[List[float], List[List[float]], Dict[str, Any]]): Embeddings or error information

        Raises:
            - KeyError: If required parameters are missing
            - TypeError: If the type of the parameter is incorrect

        Warns:
            - UserWarning: If the model is not supported or if the model is not found in the model provider
        """

        # Check if the model name is provided
        if kwds.get("model_name"):
            if self.model_provider == "SentenceTransformers":
                if kwds.get("model_name") != self.model_name:
                    warnings.warn(
                        f"\nFor using different sentence-transformers model: {kwds.get('model_name')}\n"
                        f"again reinitialize the EmbeddingModels class as currently {self.model_name} is already loaded "
                        f"Hence defaulting the model to {self.model_name}",
                        UserWarning,
                    )
                # Defaulting to the model_name that is already loaded
                model_name = self.model_name
            # Check if the model is supported in the model type class
            elif kwds.get("model_name") in self.supported_models.get(
                self.model_provider
            ):
                model_name = kwds.get("model_name")
            else:
                warnings.warn(
                    f"\nModel {kwds.get('model_name')} for provider {self.model_provider} is not supported by Orichain. Supported models for {self.model_provider} are: [{', '.join(self.supported_models.get(self.model_provider))}] \nUsing an unsupported model may lead to unexpected issues. Please verify that you are using the correct 'model_name'.",
                    UserWarning,
                )
                # Proceeding with the given model name regardless
                model_name = kwds.get("model_name")
            kwds.pop("model_name")
        else:
            model_name = self.model_name

        # Get the embeddings
        user_message_vector = self.model(
            text=user_message, model_name=model_name, **kwds
        )

        return user_message_vector


class AsyncEmbeddingModel(object):
    """Asynchronus Base class for embedding generation.

    This class provides a unified interface to interact with different embedding models from providers such as OpenAI, AWS Bedrock, Google Gemini and Vertex AI, Azure OpenAI, and SentenceTransformers."""

    default_model = DEFUALT_EMBEDDING_MODEL
    default_model_provider = DEFAULT_MODEL_PROVIDER
    supported_models = SUPPORTED_MODELS
    model_handler = {
        "OpenAI": openai_embeddings.AsyncEmbed,
        "AWSBedrock": awsbedrock_embeddings.AsyncEmbed,
        "SentenceTransformers": stransformers_embeddings.AsyncEmbed,
        "AzureOpenAI": azureopenai_embeddings.AsyncEmbed,
        "GoogleGemini": gcp_gemini_embeddings.AsyncEmbed,
        "GoogleVertexAI": gcp_vertex_embeddings.AsyncEmbed,
        "TogetherAI": togetherai_embeddings.AsyncEmbed,
    }

    def __init__(self, **kwds: Any) -> None:
        """Initialize the Embedding Models class with the required parameters.

        Args:
            - model_name (str, optional): Name of the model to be used. Default: "text-embedding-3-small"
            - provider (str, optional): Name of the model provider. Default: "OpenAI". Allowed values:
                - OpenAI
                - AWSBedrock
                - GoogleGemini
                - GoogleVertexAI
                - AzureOpenAI
                - TogetherAI
                - SentenceTransformers

            **Authentication Arguments by provider:**

                **OpenAI models:**
                    - api_key (str): OpenAI API key.
                    - timeout (Timeout, optional): Request timeout parameter like connect, read, write. Default: 60.0, 5.0, 10.0, 2.0
                    - max_retries (int, optional): Number of retries for the request. Default: 2

                **AWS Bedrock models:**
                    - aws_access_key (str): AWS access key.
                    - aws_secret_key (str): AWS secret key.
                    - aws_region (str): AWS region name.
                    - config (Config, optional):
                        - connect_timeout (float or int, optional): The time in seconds till a timeout exception is thrown when attempting to make a connection. Default: 60
                        - read_timeout: (float or int, optional): The time in seconds till a timeout exception is thrown when attempting to read from a connection. Default: 60
                        - region_name (str, optional): region name Note: If specifing config you need to still pass region_name even if you have already passed in aws_region
                        - max_pool_connections: The maximum number of connections to keep in a connection pool. Defualt: 10
                        - retries (Dict, optional):
                            - total_max_attempts: Number of retries for the request. Default: 2

                **Google Gemini models:**
                    - api_key (str): Gemini API key
                    - http_options (types.HttpOptions, optional): HTTP options to be used in each of the requests. Default is None
                    - debug_config (DebugConfig, optional): Configuration options that change client network behavior when testing. Default is None

                **Google Vertex AI models:**
                    - api_key (str): Vertex AI API key
                    - credentials (google.auth.credentials.Credentials): The credentials to use for authentication when calling the Vertex AI APIs.
                    - project (str): The Google Cloud project ID to use for quota.
                    - location (str): The location to send API requests to (for example, us-central1).
                    - http_options (types.HttpOptions, optional): HTTP options to be used in each of the requests. Default is None
                    - debug_config (DebugConfig, optional): Configuration options that change client network behavior when testing. Default is None

                **Sentence Transformers models:**
                    - model_download_path (str, optional): Path to download the model. Default: "/home/ubuntu/projects/models/embedding_models"
                    - device (str, optional): Device to run the model. Default: "cpu"
                    - trust_remote_code (bool, optional): Trust remote code. Default: False
                    - token (str, optional): Hugging Face API token

                **Azure OpenAI models:**
                    - api_key (str): Azure OpenAI API key.
                    - azure_endpoint (str): Azure OpenAI endpoint.
                    - api_version (str): Azure OpenAI API version.
                    - timeout (Timeout, optional): Request timeout parameter like connect, read, write. Default: 60.0, 5.0, 10.0, 2.0
                    - max_retries (int, optional): Number of retries for the request. Default: 2

                **TogetherAI models:**
                    - api_key (str): TogetherAI API key.
                    - timeout (Timeout, optional): Request timeout in seconds. Default: 60
                    - max_retries (int, optional): Number of retries for the request. Default: 2

        Raises:
            - ValueError: If the model is not supported
            - KeyError: If required parameters are missing
            - TypeError: If the type of the parameter is incorrect
            - ImportError: If the required library is not installed

        Warns:
            - UserWarning: If no model_name is provided, defaulting to `text-embedding-3-small`
        """

        # Check if the model name is provided
        if not kwds.get("model_name"):
            warnings.warn(
                f"\nNo 'model_name' specified, hence defaulting to {self.default_model}",
                UserWarning,
            )
        if not kwds.get("provider"):
            warnings.warn(
                f"\nNo 'provider' specified, hence defaulting to {self.default_model_provider}",
                UserWarning,
            )

        self.model_name = kwds.get("model_name", self.default_model)
        self.model_provider = kwds.get("provider", self.default_model_provider)

        # Validating model name and model provider name
        if self.model_provider not in self.model_handler:
            raise ValueError(
                f"\nUnsupported model provider: {self.model_provider}\nSupported providers are:"
                f"\n- " + "\n- ".join(list(self.model_handler.keys()))
            )
        elif self.model_provider == "SentenceTransformers":
            repo_check = hf_repo_exists(
                repo_id=self.model_name,
                repo_type=kwds.get("repo_type"),
                token=kwds.get("token"),
            )
            if not repo_check:
                raise ValueError(
                    f"\nThe Huggingface repository '{self.model_name}' does not exist. \nPlease ensure you provide the full repository path in 'model_name'."
                )
        elif self.model_name not in self.supported_models.get(self.model_provider):
            warnings.warn(
                f"\nModel {self.model_name} for provider {self.model_provider} is not supported by Orichain. Supported models for {self.model_provider} are: [{', '.join(self.supported_models.get(self.model_provider))}] \nPlease make sure you're using the correct 'model_name' and 'provider'",
                UserWarning,
            )
        else:
            pass

        # Initialize the model
        self.model = self.model_handler.get(self.model_provider)(**kwds)

    async def __call__(
        self, user_message: Union[str, List[str]], **kwds: Any
    ) -> Union[List[float], List[List[float]], Dict]:
        """Get embeddings for the given text(s).

        Args:
            - user_message (Union[str, List[str]]): Input text or list of texts

            **Generation Arguments by provider:**

                **OpenAI & Azure OpenAI models:**
                    - model_name (str, optional): Name of the embedding model to use.

                **Google Gemini & Vertex AI models:**
                    - model_name (str, optional): Name of the embedding model to use
                    - config (google.genai.types.EmbedContentConfig, optional): Optional model configuration parameters provided to the client.models.embed_content API.

                **AWS Bedrock models:**
                    - model_name (str, optional): Name of the embedding model to use.

                    **Cohere Embedding Models:**
                        - input_type (Literal["search_query", "search_document", "classification", "clustering", "image"], optional): Type of input text. Default: "search_query"
                        - embedding_types (str, optional): Specifies the types of embeddings you want to have returned. Can be one or more of the following types: 'float', 'int8', 'uint8', 'binary', 'ubinary'
                        - truncate (Literal["NONE", "START", "END"], optional): Specifies how the API handles inputs longer than the maximum token length. Use one of the following:
                            - NONE – (Default) Returns an error when the input exceeds the maximum input token length.
                            - START – Discards the start of the input.
                            - END – Discards the end of the input.

                    **Amazon Titan Embeddings G1 Models:**
                        - dimensions (int, optional): Output dimensions. Default: 1024 (Output dimensions can be: 256, 512 and 1024)
                        - normalize (bool, optional): Normalize the output. Default: True (As recommended in docs for RAG)

                **Sentence Transformers models:**
                    - prompt_name (str, optional): The name of the prompt to use for encoding. Must be a key in the `prompts` dictionary,
                      which is either set in the constructor or loaded from the model configuration. For example if
                      ``prompt_name`` is "query" and the ``prompts`` is {"query": "query: ", ...}, then the sentence "What
                      is the capital of France?" will be encoded as "query: What is the capital of France?" because the sentence
                      is appended to the prompt. If ``prompt`` is also set, this argument is ignored. Defaults to None.
                    - prompt (str, optional): The prompt to use for encoding. For example, if the prompt is "query: ", then the
                      sentence "What is the capital of France?" will be encoded as "query: What is the capital of France?"
                      because the sentence is appended to the prompt. If ``prompt`` is set, ``prompt_name`` is ignored. Defaults to None.
                    - output_value (Literal["sentence_embedding", "token_embeddings"], optional): The type of embeddings to return:
                      "sentence_embedding" to get sentence embeddings, "token_embeddings" to get wordpiece token embeddings, and `None`,
                      to get all output values. Defaults to "sentence_embedding".
                    - show_progress_bar (bool, optional): Whether to output a progress bar when encode sentences. Defaults to False.
                    - precision (Literal["float32", "int8", "uint8", "binary", "ubinary"], optional): The precision to use for the embeddings.
                      Can be "float32", "int8", "uint8", "binary", or "ubinary". All non-float32 precisions are quantized embeddings.
                      Quantized embeddings are smaller in size and faster to compute, but may have a lower accuracy. They are useful for
                      reducing the size of the embeddings of a corpus for semantic search, among other tasks. Defaults to "float32".
                    - batch_size (int, optional): The batch size used for the computation. Defaults to 32.
                    - convert_to_numpy (bool, optional): Whether the output should be a list of numpy vectors. If False, it is a list of PyTorch tensors.
                      Defaults to False.
                    - convert_to_tensor (bool, optional): Whether the output should be one large tensor. Overwrites `convert_to_numpy`.
                      Defaults to False.
                    - device (str, optional): Which :class:`torch.device` to use for the computation. Defaults to None.
                    - normalize_embeddings (bool, optional): Whether to normalize returned vectors to have length 1. In that case,
                      the faster dot-product (util.dot_score) instead of cosine similarity can be used. Defaults to False.

        Returns:
            (Union[List[float], List[List[float]], Dict[str, Any]]): Embeddings or error information

        Raises:
            - KeyError: If required parameters are missing
            - TypeError: If the type of the parameter is incorrect

        Warns:
            - UserWarning: If the model is not supported or if the model is not found in the model provider
        """

        # Check if the model name is provided
        if kwds.get("model_name"):
            if self.model_provider == "SentenceTransformers":
                if kwds.get("model_name") != self.model_name:
                    warnings.warn(
                        f"\nFor using different sentence-transformers model: {kwds.get('model_name')}\n"
                        f"again reinitialize the EmbeddingModels class as currently {self.model_name} is already loaded "
                        f"Hence defaulting the model to {self.model_name}",
                        UserWarning,
                    )
                # Defaulting to the model_name that is already loaded
                model_name = self.model_name
            # Check if the model is supported in the model type class
            elif kwds.get("model_name") in self.supported_models.get(
                self.model_provider
            ):
                model_name = kwds.get("model_name")
            else:
                warnings.warn(
                    f"\nModel {kwds.get('model_name')} for provider {self.model_provider} is not supported by Orichain. Supported models for {self.model_provider} are: [{', '.join(self.supported_models.get(self.model_provider))}] \nUsing an unsupported model may lead to unexpected issues. Please verify that you are using the correct 'model_name'.",
                    UserWarning,
                )
                # Proceeding with the given model name regardless
                model_name = kwds.get("model_name")
            kwds.pop("model_name")
        else:
            model_name = self.model_name

        # Get the embeddings
        user_message_vector = await self.model(
            text=user_message, model_name=model_name, **kwds
        )

        return user_message_vector



import os
import sys

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Orichain"
copyright = "2025, Oriserve"
author = "Apoorv Singh, Shubham Maindola"
release = "2.4.0"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = []
templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_show_sourcelink = False
html_theme = "furo"
html_static_path = ["_static"]
html_css_files = ["custom.css"]
html_theme_options = {
    "light_logo": "logo-light.png",
    "dark_logo": "logo-dark.png",
}


# -- Enabling sphinx.ext.autodoc (and others) to pull docstrings automatically -------------------------------------------------

extensions = ["sphinx.ext.autodoc", "sphinx.ext.viewcode"]
sys.path.insert(0, os.path.abspath("../../src"))
autodoc_member_order = "bysource"



from typing import Any, List, Union, Optional, Dict
import warnings

from orichain.knowledge_base import pinecone_knowledgbase, chromadb_knowledgebase
from orichain import error_explainer

DEFAULT_KNOWLEDGE_BASE = "pinecone"


class KnowledgeBase(object):
    """
    Synchronous interface for interacting with vector databases.

    This class provides a unified API to communicate with supported vector databases.
    Currently, Pinecone and ChromaDB are supported.
    """

    default_knowledge_base = DEFAULT_KNOWLEDGE_BASE

    def __init__(self, vector_db_type: Optional[str], **kwds: Any) -> None:
        """Initializes the knowledge base.

        Args:
            - vector_db_type (str, optional): Type of knowledge base. Default: pinecone

        **Authentication parameters by provider:**

            **Pinecone:**
                - api_key (str): Pinecone API key
                - index_name (str): Pinecone index name
                - namespace (str): Pinecone namespace

            **ChromaDB:**
                - collection_name (str): ChromaDB collection name
                - path (str, optional): Path to the ChromaDB database Default: `/home/ubuntu/projects/chromadb`

        Raises:
            - ValueError: If the knowledge base type is not supported
            - KeyError: If the required params is not found

        Warns:
            - UserWarning: If the knowledge base type is not defined Default: pinecone
        """
        try:
            # Dictionary mapping vector database types to their respective handler classes
            knowledge_base_handler = {
                "pinecone": pinecone_knowledgbase.DataBase,
                "chromadb": chromadb_knowledgebase.DataBase,
            }

            # If no vector_db_type is provided, default to pinecone
            if not vector_db_type:
                warnings.warn(
                    f"\nKnowledge base type not defined hence defaulting to \
                    {self.default_knowledge_base}",
                    UserWarning,
                )
                self.vector_db_type = self.default_knowledge_base
            # If vector_db_type is not supported, raise a ValueError
            elif vector_db_type not in list(knowledge_base_handler.keys()):
                raise ValueError(
                    f"\nUnsupported knowledge base: {self.model_name}\nSupported knowledge bases are:"
                    f"\n- " + "\n- ".join(list(knowledge_base_handler.keys()))
                )
            else:
                self.vector_db_type = vector_db_type

            # Initialize the knowledge base handler
            self.retriver = knowledge_base_handler.get(
                vector_db_type, self.default_knowledge_base
            )(**kwds)

        except Exception as e:
            error_explainer(e)

    def __call__(
        self,
        num_of_chunks: int,
        user_message_vector: Optional[List[Union[int, float]]] = None,
        **kwds: Any,
    ) -> Dict:
        """Retrieves the chunks from the knowledge base

        Args:
            - user_message_vector (Optional[List[Union[int, float]]]): Embedding of the text. Defaults to None.
            - num_of_chunks (int): Number of chunks to retrieve

            **Retrieval Arguments by VectorDB:**

                **Pinecone:**
                    - vector (List[float]): The query vector. This should be the same length as the dimension of the index
                      being queried. Each `query()` request can contain only one of the parameters `id` or `vector`.. [optional]
                    - id (str): The unique ID of the vector to be used as a query vector.
                      Each `query()` request can contain only one of the parameters `vector` or  `id`.. [optional]
                    - top_k (int): The number of results to return for each query. Must be an integer greater than 1.
                    - namespace (str): The namespace to fetch vectors from.
                      If not specified, the default namespace is used. [optional]
                    - filter (Dict[str, Union[str, float, int, bool, List, dict]]):
                      The filter to apply. You can use vector metadata to limit your search.
                      See https://www.pinecone.io/docs/metadata-filtering/ [optional]
                    - include_values (bool): Indicates whether vector values are included in the response.
                      If omitted the server will use the default value of False [optional]
                    - include_metadata (bool): Indicates whether metadata is included in the response as well as the ids.
                      If omitted the server will use the default value of False  [optional]
                    - sparse_vector: (Union[SparseValues, Dict[str, Union[List[float], List[int]]]]): sparse values of the query vector.
                      Expected to be either a SparseValues object or a dict of the form:
                      ``{'indices': List[int], 'values': List[float]}``, where the lists each have the same length.

                **ChromaDB:**
                    - collection_name (str, optional): The name of the collection to get documents from. Defaults to the collection_name set during class instantiation.
                    - where (Dict, optional):  A Where type dict used to filter results by. E.g. ``{$and: [{"color" : "red"}, {"price": 4.20}]}``. Default: None.
                    - where_document (Dict, optional): A WhereDocument type dict used to filter by the documents. E.g. ``{"$contains" : "hello"}``. Default: None.
                    - include (List, optional): A list of what to include in the results.
                      Can contain ``"embeddings"``, ``"metadatas"``, ``"documents"``, ``"distances"``.
                      Ids are always included. Defaults to ``["metadatas", "documents", "distances"]``.
                      Default: ``["metadatas", "documents"]``

        Returns:
            Dict: Result of retrieving the chunks

        Raises:
            - ValueError: If `user_message_vector` is needed except for pinecone but if ids are also not provided for pinecone this error will be raised
            - KeyError: If required `namespace` is not found for pinecone
        """
        try:
            if not user_message_vector and not self.vector_db_type == "pinecone":
                raise ValueError("`user_message_vector` is needed except for pinecone")

            chunks = self.retriver(
                user_message_vector=user_message_vector,
                num_of_chunks=num_of_chunks,
                **kwds,
            )

            return chunks

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}

    def fetch(
        self,
        ids: List[str],
        **kwds: Any,
    ) -> Dict:
        """Fetches the chunks based on the ids from the knowledge base

        Args:
            - ids (List[str]): List of ids to fetch

            **Retrieval Arguments by VectorDB:**

                **Pinecone:**
                    - namespace (str, optional): The namespace to fetch vectors from. If not specified, the default namespace is used.

                **ChromaDB:**
                    - collection_name (str, optional): The name of the collection to fetch documents from. Defaults to the collection_name set during class instantiation.
                    - limit (int, optional): The number of documents to return. Default: None.
                    - offset (int, optional): The offset to start returning results from. Useful for paging results with limit. Default: None.
                    - where (Dict, optional):  A Where type dict used to filter results by. E.g. ``{$and: [{"color" : "red"}, {"price": 4.20}]}``. Default: None.
                    - where_document (Dict, optional): A WhereDocument type dict used to filter by the documents. E.g. ``{"$contains" : "hello"}``. Default: None.
                    - include (List, optional): A list of what to include in the results.
                      Can contain ``"embeddings"``, ``"metadatas"``, ``"documents"``, ``"distances"``.
                      Ids are always included. Defaults to ``["metadatas", "documents", "distances"]``.
                      Default: ``["metadatas", "documents"]``

        Returns:
            Dict: Result of fetching the chunks
        """
        try:
            # Fetching the chunks based on the ids
            chunks = self.retriver.fetch(
                ids=ids,
                **kwds,
            )

            return chunks

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}


class AsyncKnowledgeBase(object):
    """
    Asynchronous interface for interacting with vector databases.

    This class provides a unified API to communicate with supported vector databases.
    Currently, Pinecone and ChromaDB are supported.
    """

    default_knowledge_base = DEFAULT_KNOWLEDGE_BASE

    def __init__(self, vector_db_type: Optional[str], **kwds: Any) -> None:
        """Initializes the knowledge base.

        Args:
            - vector_db_type (str, optional): Type of knowledge base. Default: pinecone

        **Authentication parameters by provider:**

            **Pinecone:**
                - api_key (str): Pinecone API key
                - index_name (str): Pinecone index name
                - namespace (str): Pinecone namespace

            **ChromaDB:**
                - collection_name (str): ChromaDB collection name
                - path (str, optional): Path to the ChromaDB database Default: `/home/ubuntu/projects/chromadb`

        Raises:
            - ValueError: If the knowledge base type is not supported
            - KeyError: If the required params is not found

        Warns:
            - UserWarning: If the knowledge base type is not defined Default: pinecone
        """
        try:
            # Dictionary mapping vector database types to their respective handler classes
            knowledge_base_handler = {
                "pinecone": pinecone_knowledgbase.AsyncDataBase,
                "chromadb": chromadb_knowledgebase.AsyncDataBase,
            }

            # If no vector_db_type is provided, default to pinecone
            if not vector_db_type:
                warnings.warn(
                    f"\nKnowledge base type not defined hence defaulting to \
                    {self.default_knowledge_base}",
                    UserWarning,
                )
                self.vector_db_type = self.default_knowledge_base
            # If vector_db_type is not supported, raise a ValueError
            elif vector_db_type not in list(knowledge_base_handler.keys()):
                raise ValueError(
                    f"\nUnsupported knowledge base: {self.model_name}\nSupported knowledge bases are:"
                    f"\n- " + "\n- ".join(list(knowledge_base_handler.keys()))
                )
            else:
                self.vector_db_type = vector_db_type

            # Initialize the knowledge base handler
            self.retriver = knowledge_base_handler.get(
                vector_db_type, self.default_knowledge_base
            )(**kwds)

        except Exception as e:
            error_explainer(e)

    async def __call__(
        self,
        num_of_chunks: int,
        user_message_vector: Optional[List[Union[int, float]]] = None,
        **kwds: Any,
    ) -> Dict:
        """Retrieves the chunks from the knowledge base

        Args:
            - num_of_chunks (int): Number of chunks to retrieve
            - user_message_vector (Optional[List[Union[int, float]]]): Embedding of text. Defaults to None.

            **Retrieval Arguments by VectorDB:**

                **Pinecone:**
                    - vector (List[float]): The query vector. This should be the same length as the dimension of the index
                      being queried. Each `query()` request can contain only one of the parameters `id` or `vector`.. [optional]
                    - id (str): The unique ID of the vector to be used as a query vector.
                      Each `query()` request can contain only one of the parameters `vector` or  `id`.. [optional]
                    - top_k (int): The number of results to return for each query. Must be an integer greater than 1.
                    - namespace (str): The namespace to fetch vectors from.
                      If not specified, the default namespace is used. [optional]
                    - filter (Dict[str, Union[str, float, int, bool, List, dict]]):
                      The filter to apply. You can use vector metadata to limit your search.
                      See https://www.pinecone.io/docs/metadata-filtering/ [optional]
                    - include_values (bool): Indicates whether vector values are included in the response.
                      If omitted the server will use the default value of False [optional]
                    - include_metadata (bool): Indicates whether metadata is included in the response as well as the ids.
                      If omitted the server will use the default value of False  [optional]
                    - sparse_vector: (Union[SparseValues, Dict[str, Union[List[float], List[int]]]]): sparse values of the query vector.
                      Expected to be either a SparseValues object or a dict of the form:
                      ``{'indices': List[int], 'values': List[float]}``, where the lists each have the same length.

                **ChromaDB:**
                    - collection_name (str, optional): The name of the collection to get documents from. Defaults to the collection_name set during class instantiation.
                    - where (Dict, optional):  A Where type dict used to filter results by. E.g. ``{$and: [{"color" : "red"}, {"price": 4.20}]}``. Default: None.
                    - where_document (Dict, optional): A WhereDocument type dict used to filter by the documents. E.g. ``{"$contains" : "hello"}``. Default: None.
                    - include (List, optional): A list of what to include in the results.
                      Can contain ``"embeddings"``, ``"metadatas"``, ``"documents"``, ``"distances"``.
                      Ids are always included. Defaults to ``["metadatas", "documents", "distances"]``.
                      Default: ``["metadatas", "documents"]``

        Returns:
            Dict: Result of retrieving the chunks

        Raises:
            - ValueError: If `user_message_vector` is needed except for pinecone but if ids are also not provided for pinecone this error will be raised
            - KeyError: If required `namespace` is not found for pinecone
        """
        try:
            if not user_message_vector and not self.vector_db_type == "pinecone":
                raise ValueError("`user_message_vector` is needed except for pinecone")

            chunks = await self.retriver(
                user_message_vector=user_message_vector,
                num_of_chunks=num_of_chunks,
                **kwds,
            )

            return chunks

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}

    async def fetch(
        self,
        ids: List[str],
        **kwds: Any,
    ) -> Dict:
        """Fetches the chunks based on the ids from the knowledge base

        Args:
            - ids (List[str]): List of ids to fetch

            **Retrieval Arguments by VectorDB:**

                **Pinecone:**
                    - namespace (str, optional): The namespace to fetch vectors from. If not specified, the default namespace is used.

                **ChromaDB:**
                    - collection_name (str, optional): The name of the collection to fetch documents from. Defaults to the collection_name set during class instantiation.
                    - limit (int, optional): The number of documents to return. Default: None.
                    - offset (int, optional): The offset to start returning results from. Useful for paging results with limit. Default: None.
                    - where (Dict, optional):  A Where type dict used to filter results by. E.g. ``{$and: [{"color" : "red"}, {"price": 4.20}]}``. Default: None.
                    - where_document (Dict, optional): A WhereDocument type dict used to filter by the documents. E.g. ``{"$contains" : "hello"}``. Default: None.
                    - include (List, optional): A list of what to include in the results.
                      Can contain ``"embeddings"``, ``"metadatas"``, ``"documents"``, ``"distances"``.
                      Ids are always included. Defaults to ``["metadatas", "documents", "distances"]``.
                      Default: ``["metadatas", "documents"]``

        Returns:
            Dict: Result of fetching the chunks
        """
        try:
            # Fetching the chunks based on the ids
            chunks = await self.retriver.fetch(
                ids=ids,
                **kwds,
            )

            return chunks

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}



from typing import Any, List, Dict, Union
from orichain import error_explainer


class Embed(object):
    """
    Synchronous Embed class to get embeddings from TogetherAI API.
    """

    def __init__(self, **kwds: Any) -> None:
        """
        Initialize TogetherAI client and set up API key.

        Args:
            - api_key (str): TogetherAI API key
            - timeout (Timeout, optional): Request timeout. Default is 60.0
            - max_retries (int, optional): Number of retries for the request. Default is 2

        Raises:
            - KeyError: If required parameters are not provided.
            - TypeError: If an invalid type is provided for a parameter
        """

        # Validate input parameters
        if not kwds.get("api_key"):
            raise KeyError("Required `api_key` not found")
        elif kwds.get("timeout") and not isinstance(kwds.get("timeout"), (int, float)):
            raise TypeError(
                "Invalid 'timeout' type detected:",
                type(kwds.get("timeout")),
                ", Please enter valid timeout (in seconds) in either int or float.",
            )
        elif kwds.get("max_retries") and not isinstance(kwds.get("max_retries"), int):
            raise TypeError(
                "Invalid 'max_retries' type detected:,",
                type(kwds.get("max_retries")),
                ", Please enter a value that is 'int'",
            )
        else:
            pass

        # Initialize the TogetherAI client with provided parameters
        from together import Together

        self.client = Together(
            api_key=kwds.get("api_key"),
            timeout=kwds.get("timeout", 60.0),
            max_retries=kwds.get("max_retries", 2),
        )

    def __call__(
        self, text: Union[str, List[str]], model_name: str, **kwds: Any
    ) -> Union[List[float], List[List[float]], Dict]:
        """
        Get embeddings for the given text(s).

        Args:
            - text (Union[str, List[str]]): Input text or list of texts
            - model_name (str): Name of the embedding model to use
            - **kwargs: Additional keyword arguments for the embedding API

        Returns:
            (Union[List[float], List[List[float]], Dict[str, Any]]): Embeddings or error information
        """
        try:
            if isinstance(text, str):
                text = [text]

            response = self.client.embeddings.create(input=text, model=model_name)
            embeddings = [sentence.embedding for sentence in response.data]

            if len(embeddings) == 1:
                embeddings = embeddings[0]

            return embeddings
        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}


class AsyncEmbed(object):
    """
    Asynchronous Embed class to get embeddings from TogetherAI API.
    """

    def __init__(self, **kwds: Any) -> None:
        """
        Initialize TogetherAI client and set up API key.

        Args:
            - api_key (str): TogetherAI API key
            - timeout (Timeout, optional): Request timeout. Default is 60.0
            - max_retries (int, optional): Number of retries for the request. Default is 2

        Raises:
            - KeyError: If required parameters are not provided.
            - TypeError: If an invalid type is provided for a parameter
        """

        # Validate input parameters
        if not kwds.get("api_key"):
            raise KeyError("Required `api_key` not found")
        elif kwds.get("timeout") and not isinstance(kwds.get("timeout"), (int, float)):
            raise TypeError(
                "Invalid 'timeout' type detected:",
                type(kwds.get("timeout")),
                ", Please enter valid timeout (in seconds) in either int or float.",
            )
        elif kwds.get("max_retries") and not isinstance(kwds.get("max_retries"), int):
            raise TypeError(
                "Invalid 'max_retries' type detected:,",
                type(kwds.get("max_retries")),
                ", Please enter a value that is 'int'",
            )
        else:
            pass

        # Initialize the TogetherAI client with provided parameters
        from together import AsyncTogether

        self.client = AsyncTogether(
            api_key=kwds.get("api_key"),
            timeout=kwds.get("timeout", 60.0),
            max_retries=kwds.get("max_retries", 2),
        )

    async def __call__(
        self, text: Union[str, List[str]], model_name: str, **kwds: Any
    ) -> Union[List[float], List[List[float]], Dict]:
        """
        Get embeddings for the given text(s).

        Args:
            - text (Union[str, List[str]]): Input text or list of texts
            - model_name (str): Name of the embedding model to use
            - **kwargs: Additional keyword arguments for the embedding API

        Returns:
            (Union[List[float], List[List[float]], Dict[str, Any]]): Embeddings or error information
        """
        try:
            if isinstance(text, str):
                text = [text]

            response = await self.client.embeddings.create(input=text, model=model_name)
            embeddings = [sentence.embedding for sentence in response.data]

            if len(embeddings) == 1:
                embeddings = embeddings[0]

            return embeddings
        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}
